""" Module utils/io/images2d.py 
Contains common utilities specifor 2D images.

TODO:
 - Cleanup API (combined 2 old utilities files) & format inconsistencies. 
 - Use to_np inside of get_info() (done manually for now)
 - to_file tensor - normalize? unsure if differentiates btwn L and RGB
"""

import sys, os
import shutil
import math
import time
import SimpleITK as sitk
import numpy as np
from pathlib import Path
from collections import Counter
from tqdm import tqdm

import cv2
import numpy as np
import PIL.Image as Image, PIL.ImageStat as ImageStat
from skimage.transform import resize as resize
import torch
from torchvision import transforms
import pandas as pd

from . import files

### Constants ###
IMAGE_EXTS = ['.bmp', '.png', '.jpg', '.jpeg', '.tif', '.tiff', '.gif',
              '.nii', '.nii.gz', '.dcm']
CONTAINER_TYPES = ['np', 'pil', 'tensor']
COLOR_MODES = ['auto', 'rgb', 'gray', 'binary']
COLOR_2_PILMODE = {
    'rgb': 'RGB', 'gray': 'L', 'binary': '1'
}
PILMODE_2_NUMCHANNELS = {  # Note: GIFs (P) are handled as RGB by default
    '1': 1, 'L': 1, 'I': 1, 'F': 1,   # includes 1 bit, 8 bit, 32 bit pix
    'RGB': 3, 'YCbCr': 3, 'LAB': 3, 'HSV': 3, 'P': 3, # all 3x8-bit pixels
    'RGBA': 4, 'CMYK': 4  # all 4x8-bit pixels
}


### ======================================================================== ###
### * ### * ### * ### *     Module API Implementation    * ### * ### * ### * ###
### ======================================================================== ###

__all__ = ['create_dirs_from_file',
           'is_image', 
           'list_images', 'list_files', 'list_dirs', 'list_all_files_recursive',
           'get_ext', 'get_filename',
           'natural_sort']

# # # # Primary API Functions # # # #
# Note: Generally, imgobj for these can be paths, np, pil, or tensors
#       Also, visualize and get_info only ones that can take in 4D imgobjs


### ---- ### ---- \\    Image Read/Write     // ---- ### ---- ###


def read_sitk2d_gray(im_path, pixtype=sitk.sitkUInt8):
    image = sitk.ReadImage(im_path, pixtype)
    dims = image.GetNumberOfComponentsPerPixel()
    return image


def read_sitk2d_rgb(im_path, pixtype=sitk.sitkVectorUInt8):
    image = sitk.ReadImage(im_path, pixtype)
    dims = image.GetNumberOfComponentsPerPixel()
    return image


def write_np_gray2d(np_im, path, extra_channel=False):
    """ Save 2D grayscale image of shape 1xHxW, HxW, or HxWx1
    """
    if extra_channel:
        np_im = np_im.squeeze(0).squeeze(-1)
    assert np_im.ndim == 2, f"Only 2D gray images are supported."
    img = sitk.GetImageFromArray(np_im, isVector=False)
    _write_image(img, path)


def write_np_rgb2d(np_im, path, channel_last=True):
    """ Save 2D grayscale image of shape 1xHxW, HxW, or HxWx1
    """
    if not channel_last:
        np_im = to_np_channel_last(np_im)
    assert np_im.ndim == 3 and np_im.shape[-1] == 3, \
        f"Only 2D RGB images are supported."
    img = sitk.GetImageFromArray(np_im, isVector=True)
    _write_image(img, path)


def write_np_rgba2d(np_im, path, channel_last=True):
    if channel_last:
        np_im = to_np_channel_first(np_im)
    C = np_im.shape[0]
    channels = [sitk.GetImageFromArray(np_im[i, :, :]) for i in range(C)]
    filter = sitk.ComposeImageFilter()
    img = filter.Execute(*[channels[i] for i in range(C)])
    _write_image(img, path)


def _write_image(img, path):
    r""" Creates directory structure and writes using sitk's writer. """ 
    file_utils.create_dirs_from_file(path)
    try:
        writer = sitk.ImageFileWriter()
        writer.SetFileName(path)
        writer.SetUseCompression(compress)
        writer.Execute(img)
        return True
    except:
        print(f"Error writing image to {path}!")
        return False
    
    
### ---- ### ---- \\    Image Transforms     // ---- ### ---- ###

def to_np_channel_last(np_im):
    return np.moveaxis(np_im, 0, -1)


def to_np_channel_first(np_im):
    return np.moveaxis(np_im, -1, 0)


def to_np(imgobj, size=None, color=None, extra_channel=False):
    """
    (1) np:     RGB HxWx3, Gray/Bin HxW
    (2) pil:    RGB 'RGB', Gray 'L', Bin '1'
    (3) tensor: RGB 3xHxW, Gray/bin HxW

    is_single_channel = False  # assumes RGB
    if color == 'auto':
        img = Image.open(filepath).convert('RGB')
        is_single_channel = is_grayscale(img)
    elif 'bin' in color or 'gray' in color:
        is_single_channel = True
    """
    img = None
    if isinstance(imgobj, str):
        img = _read_img(imgobj, container='np')
    elif isinstance(imgobj, np.ndarray):  
        img = imgobj
    elif 'PIL.' in str(type(imgobj)):  
        img = np.uint8(np.array(imgobj))
    elif isinstance(imgobj, torch.Tensor):
        if torch.min(imgobj) >= 0 and torch.max(imgobj) <= 1:
            imgobj *= 255.
        elif torch.min(imgobj) < 0 or torch.max(imgobj) > 255.:
            print(f"to_np WARNING: {imgobj.shape} tensor has min ({torch.min(imgobj)}) and max ({torch.max(imgobj)})")
        if len(imgobj.shape) == 3:
            if imgobj.shape[0] == 1 or imgobj.shape[0] == 3:
                imgobj = imgobj.permute(1,2,0)
        img = np.uint8(imgobj.cpu().numpy())
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")
    
    if size:
        img = _resize(img, size)
    if color:
        img = _recolor(img, color, extra_channel=extra_channel)
    return img


def to_tensor(imgobj, color=None, size=None,
              extra_channel=False):
    
    img = None
    if isinstance(imgobj, str):
        img = _read_img(imgobj, container='ten')
    elif isinstance(imgobj, np.ndarray):
        img = torch.from_numpy(imgobj).float()/255.
        if img.ndim == 3:
            img = img.permute(2,0,1)
    elif 'PIL.' in str(type(imgobj)):
        if imgobj.mode == 'L':
            img = transforms.ToTensor()(imgobj)[0,:,:]  # adds extra dim when L
        else:
            img = transforms.ToTensor()(imgobj)
    elif isinstance(imgobj, torch.Tensor):
        img = imgobj
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")
    
    if size:
        img = _resize(img, size)
    if color:
        img = _recolor(img, color, extra_channel=extra_channel)
    return img


def to_pil(imgobj, color=None, size=None):
    
    img = None
    if isinstance(imgobj, str):
        img = _read_img(imgobj, container='pil')
    elif isinstance(imgobj, np.ndarray):
        if imgobj.ndim == 3 and imgobj.shape[-1] == 1:  # HxWx1 not supported
            img = Image.fromarray(imgobj[...,0]).convert('L')  
        else:
            if imgobj.ndim == 3:
                img = Image.fromarray(imgobj).convert('RGB')
            else:
                img = Image.fromarray(imgobj).convert('L')
    elif 'PIL.' in str(type(imgobj)):
        img = imgobj
    elif isinstance(imgobj, torch.Tensor):
        imgobj = to_np(imgobj)
        if imgobj.ndim == 3 and imgobj.shape[-1] == 1:  # HxWx1 not supported
            img = Image.fromarray(imgobj[...,0]).convert('L')  
        else:
            if imgobj.ndim == 3:
                img = Image.fromarray(imgobj).convert('RGB')
            else:
                img = Image.fromarray(imgobj).convert('L')
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")
    
    if size:
        img = _resize(img, size)
    if color:
        img = _recolor(img, color)
    return img


def to_file(imgobj, path, color=None, size=None, override=False, mkdir=True):
    r"""
    Saves imgobj to path (image file).

    Parameters:
        imgobj - np array, pil obj, tensor representing a single image
        path - full path including the image name and extension
    """
    # Check assumptions: save dir is valid, no file exists already, etc.
    if os.path.isfile(path):
        if override:
            print(f"(to_file) overriding file '{path}'")
        else:
            raise IOError(f"File ({path}) already exists & override is off")
    if not os.path.isdir(os.path.dirname(path)):
        if mkdir:
            print(f"(to_file) making directory for image's path "
                  f"'{os.path.dirname(path)}'")
            Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
        else:
            raise IOError(f"Directory ({os.path.dirname(path)}) does not "
                          f"exist and mkdir is set to False.")
    
    img = None  # get input img as pil object
    if isinstance(imgobj, str):
        assert files.is_image(imgobj), f'Invalid image file given ({imgobj})'
        if files.get_ext(imgobj) == files.get_ext(path):
            shutil.copy(imgobj, path)  # overrides by default, checks above
            return
        else:  # image format is different, needs conversion
            img = Image.open(imgobj)
    elif isinstance(imgobj, np.ndarray):
        dims = imgobj.shape
        assert len(imgobj.shape) == 2 or len(imgobj.shape) == 3
        if len(dims) == 3:
            assert dims[-1] == 1 or dims[-1] == 3
            if dims[-1] == 1:
                img = Image.fromarray(imgobj).convert('L')
            else:
                img = Image.fromarray(imgobj).convert('RGB')
        else:  # 2D grayscale HxW image
            img = Image.fromarray(imgobj).convert('L') # assumed range: 0 to 255
    elif 'PIL.' in str(type(imgobj)):
        img = imgobj
    elif isinstance(imgobj, torch.Tensor):
        img = transforms.ToPILImage()(imgobj)
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")

    if size:
        img = _resize(img, size)
    if color:
        img = _recolor(img, color)
    print(f"(to_file) writing image to '{path}'")
    img.save(path)
    return path


### ---- ### ---- \\   Image Queries, Listing, Properties   // ---- ### ---- ###


def get_info(imgobjs, verbose=True):
    r"""
    Print basic info of image or image stack of filepath, np, pil or tensor.
        > Assumes images in stack or list have same number of channels.

    Parameters:
        imgobjs - individual img (str,np,pil,ten) or stack (strs,4dnp,pils,4dten)
        verbose - prints statistics for each individual image instance
        std     - display standard deviation over all images for each channel
    Returns: the dimensions as B, C, H, W.
    """

    ## Put entries in list format & figure out N
    N, formatted_imgobjs = 0, []
    display_exts = False
    if isinstance(imgobjs, list):
        for io in imgobjs:
            formatted_imgobjs += _create_img_list(io)
            if isinstance(io, str):
                display_exts = True
    else:
        formatted_imgobjs = _create_img_list(imgobjs)
    N = len(formatted_imgobjs)
    assert N > 0

    print(f"\nCollecting information from {N} images..\n", flush=True)
    if verbose and N > 1:
        print(f"- - - - - - -\nInstance Info\n- - - - - - -")

    ## Collect Info
    # Globals
    gray_counter = 0
    dims_2_count, exts_2_count = Counter(), Counter()
    sum_sizes = [0., 0.]  # H, W
    # RGB
    rgb_absmin, rgb_absmax = [0., 0., 0.], [0., 0., 0.]
    sum_rgb_minvals, sum_rgb_maxvals = [0., 0., 0.], [0., 0., 0.]
    sum_rgb_means, sum_rgb_stds = [0., 0., 0.], [0., 0., 0.]
    # gray
    gray_absmin, gray_absmax = 0., 0.
    sum_gray_minvals, sum_gray_maxvals = 0., 0.
    sum_gray_means, sum_gray_stds = 0., 0.
    
    rgb_initialized, gray_initialized = False, False

    iterable = formatted_imgobjs if verbose else tqdm(formatted_imgobjs)
    for i, img in enumerate(iterable):
        minval, maxval = [], []
        mean, std = [], []
        dim, ext = None, None
        
        np_img = to_np(img)
        
        is_gray = True
        if np_img.ndim == 3:
            is_gray = False if np_img.shape[-1] == 3 else True
            for c in range(np_img.shape[-1]):
                minval.append(np.min(np_img[:,:,c]))
                maxval.append(np.max(np_img[:,:,c]))
                mean.append(np.mean(np_img[:,:,c]))
                std.append(np.std(np_img[:,:,c]))
        else:
            minval.append(np.min(np_img[:,:]))
            maxval.append(np.max(np_img[:,:]))
            mean.append(np.mean(np_img[:,:]))
            std.append(np.std(np_img[:,:]))

                
        
        # Update globals for current image
        if is_gray: 
            gray_counter += 1
            if not gray_initialized:
                gray_initialized = True
                gray_absmin = minval[0]
                gray_absmax = maxval[0]
            else:
                if minval[0] < gray_absmin:
                    gray_absmin = minval[0]
                if maxval[0] > gray_absmax:
                    gray_absmax = maxval[0]
            sum_gray_minvals += minval[0]
            sum_gray_maxvals += maxval[0]
            sum_gray_means += mean[0]
            sum_gray_stds += std[0]
        else:  # is rgb
            if not rgb_initialized:
                rgb_initialized = True
                rgb_absmin = minval
                rgb_absmax = maxval
            else:
                for c in range(len(minval)):
                    if minval[c] < rgb_absmin[c]:
                        rgb_absmin[c] = minval[c]
                    if maxval[c] > rgb_absmax[c]:
                        rgb_absmax[c] = maxval[c]
            for c in range(len(minval)):
                sum_rgb_minvals[c] += minval[c]
                sum_rgb_maxvals[c] += maxval[c]
                sum_rgb_means[c] += mean[c]
                sum_rgb_stds[c] += std[c]
        dim = get_dimensions(np_img)
        dims_2_count[dim] += 1
        sum_sizes[0] += dim[2]
        sum_sizes[1] += dim[3]
        if isinstance(img, str):
            ext = files.get_ext(img)
            exts_2_count[ext] += 1
        
        # print individual instance info
        if verbose and N > 1:
            print(f"> Image #{i+1}/{N}. Size: {dim}.")
            if isinstance(img, str):
                print(f"\tFile: {files.get_filename(img)}")
            print(f"\tMeans:  {mean} \n\tStds:   {std}\n"
                  f"\tMins:   {minval} \n\tMaxes:  {maxval}")
    
    ## Print Global Info
    N_RGB = N - gray_counter

    print(f"\n- - - - - -\nGlobal Info\n- - - - - -")
    print(f"Num Images:  {N} ({N-gray_counter} RGB, {gray_counter} Grayscale)")
    print(f"Average Size: {int(sum_sizes[0]/N)}x{int(sum_sizes[1]/N)}")
    if display_exts:
        print(f"\tFile Types & Counts..")
        for counts in exts_2_count.most_common(len(exts_2_count)):
            print(f"\t\t{counts[0]}\t{counts[1]}")
    print(f"\tImage Sizes & Counts..")
    for counts in dims_2_count.most_common(len(dims_2_count)):
        print(f"\t\t{counts[0]}\t{counts[1]}")

    if N_RGB > 0:
        print(f"\n[RGB Images]")
        print(f"Avg Maxes:  {[t/N_RGB for t in sum_rgb_maxvals]}\n"
              f"Abs Maxes:  {rgb_absmax}")
        print(f"Avg Minis:  {[t/N_RGB for t in sum_rgb_minvals]}\n"
              f"Abs Minis:  {rgb_absmin}")
        print()
        rgb_means = [t/N_RGB for t in sum_rgb_means]
        print(f"Norm-Means: {rgb_means} \n"
              f"Abs Mean:    {sum(rgb_means)/3}")
        rgb_stds = [t/N_RGB for t in sum_rgb_stds]
        print(f"Norm-Stds:  {rgb_stds}\n"
              f"Abs Std:     {sum(rgb_stds)/3}") 
    
    if gray_counter > 0:
        print(f"\n[Gray Images]")
        print(f"Avg Max:     {sum_gray_maxvals/gray_counter}\n"
              f"Abs Max:     {gray_absmax}")
        print(f"Avg Mini:    {sum_gray_minvals/gray_counter}\n"
              f"Abs Mini:    {gray_absmin}")
        print()
        print(f"Norm-Mean:   {sum_gray_means/gray_counter} ")
        print(f"Norm-Std:    {sum_gray_stds/gray_counter} ")


def collate(imgobjs, container='np', color=None, size=None):
    r"""
    Takes in list (img objs) and collates the data into 4D stack (np/tensor).

    Parameters:
        imgobjs - list of filenames, path(s), pil objects, 
                  3d/4d tensors/np
        container - can be np stack or tensor stack from np
        color - rgb or single channel gray (Bx1xHxW)
        size - resized image (if not given, defaults to avg image size)
    """
    
    ## collects objects into list and dimensions
    imgs = []
    if isinstance(imgobjs, list):
        for io in imgobjs:
            imgs += _create_img_list(io)
    else:
        imgs = _create_img_list(imgobjs)

    ## zero-initiated placeholder stack 
    channels = 0
    if not size:
        size = [0, 0]  # H, W
        for img in imgs:
            _, C, H, W = get_dimensions(img)
            size[0] += H
            size[1] += W
            channels = C if C > channels else channels
        size = [int(d/len(imgs)) for d in size]
    assert len(size) == 2 and size[0] > 0 and size[1] > 0

    if color:
        channels = 3 if color == 'rgb' else 1
    else:
        color = 'rgb' if channels == 3 else 'gray'
    
    ## create and return stack
    if container == 'np':
        stack = np.zeros((len(imgs),size[0],size[1],channels))
        for i, img in enumerate(imgs):
            img = to_np(img, size=size, color=color, extra_channel=True)
            stack[i,...] = img
        return stack
    else:
        stack = torch.zeros((len(imgs),channels,size[0],size[1]))
        for i, img in enumerate(imgs):
            stack[i,:,:,:] = to_tensor(img, size=size, color=color)
        return stack
    

def get_dimensions(imgobjs):
    r"""  *4D stack accepted for matrix representations only*

    Returns B,C,H,W for a single image (file, np, pil, tens) or 4D (np, tens).
        Note: for file paths, this process involves MSE calculating.

    Parameters
        imgobjs - can be a filepath, pil object, np image (BxHxWxC),
                  or torch tensor (BxCxHxW)
    """

    num_images, num_channels, num_rows, num_cols = 1, 3, 0, 0
    if isinstance(imgobjs, str) or 'PIL.' in str(type(imgobjs)):
        # PIL.open ~10x faster than imread to np
        if isinstance(imgobjs, str):
            img = _read_img(imgobjs, container='pil')
        else:
            img = imgobjs
        num_cols, num_rows = img.size
        if len(img.getbands()) == 1:
            num_channels = 0
    elif isinstance(imgobjs, np.ndarray):
        assert imgobjs.ndim >= 2 and imgobjs.ndim <= 4, \
               f"Given numpy shape ({imgobjs.shape}) doesn't describe an image."
        shape = imgobjs.shape
        if len(shape) == 4:  # 4D stack
            num_images, num_rows, num_cols, num_channels = imgobjs.shape
        elif len(shape) == 3:  # assumed 1 RGB image
            assert shape[-1] == 3 or shape[-1] == 1
            num_rows, num_cols, num_channels = shape
        else:  # 2 ndims
            num_rows, num_cols = shape
            num_channels = 0
    elif isinstance(imgobjs, torch.Tensor):
        assert imgobjs.ndim >= 2 and imgobjs.ndim <= 4, \
               f"Given tensor shape ({imgobjs.shape}) doesn't describe an image."
        size = imgobjs.size()
        if len(size) == 4:  # 4D stack
            num_images, num_channels, num_rows, num_cols = size
        elif len(size) == 3:  # assumed 1 RGB image
            assert size[0] == 1 or size[0] == 3
            num_channels, num_rows, num_cols = size
        else:
            num_rows, num_cols = size
            num_channels = 0
    else:
        raise TypeError(f"Image obj of type ({type(imgobjs)}) is not valid.")

    return num_images, num_channels, num_rows, num_cols


def get_mean_std_norms(imgobjs, use_float=False):
    r"""
    Returns the average means & stds for each image channel.

    Parameters
        imgobjs - can be list (or single) of 3d/4d np/tens, pils, files/paths
        use_float - true transforms image intensities to range [0,1]
    """
    
    ## Put entries in list format & figure out N
    N, formatted_imgobjs = 0, []
    if isinstance(imgobjs, list):
        for io in imgobjs:
            formatted_imgobjs += _create_img_list(io)
    else:
        formatted_imgobjs = _create_img_list(imgobjs)
    N = len(formatted_imgobjs)
    assert N > 0

    ## calculates running sums of means/stds
    sum_means, sum_stds = [0., 0., 0.], [0., 0., 0.]
    for i, img in enumerate(formatted_imgobjs):
        mean, std = [0., 0., 0.], [0., 0., 0.]  # if gray, updates first val
        np_img = to_np(img)
        if use_float:
            np_img /= 255.
        
        if np_img.ndim == 3:
            for c in range(np_img.shape[-1]):
                mean[c] = np.mean(np_img[:,:,c])
                std[c] = np.std(np_img[:,:,c])
        else:
            mean[0] = np.mean(np_img[:,:])
            std[0] = np.std(np_img[:,:])

        for c in range(3):
            sum_means[c] += mean[c]
            sum_stds[c] += std[c]

    return [m/N for m in sum_means], [s/N for s in sum_stds]


# # # # Secondary Functions # # # #
# Note: these don't worry about image paths as imgobj

def is_grayscale(imgobj, thumbsize=40, MSEthreshold=22, adjust_color_bias=False):
    r"""  *Note: this is a slow, hard check that is data dependent!
    Checks if given image (filepath, pil, np, tensor) is grayscale.
    (1) Converts to RGB  (2) Computer average channel-variance per pixel
    
    Parameters
        imgobj - can be filepath, np, pil, tens
        thumbsize - downsized image to be analyzed (default 40x40)
        MSEthreshold - considered binary if MSE channel vals <= this
        adjust_color_bias - subtracts every value from mean value (unnecessary)
    """
    debug = False

    pil_img = imgobj if 'PIL.' in str(type(imgobj)) else to_pil(imgobj)
    if pil_img.getbands()[0] == 'P':
        pil_img = pil_img.convert('RGB')
    
    bands = pil_img.getbands()
    if bands == ('R','G','B') or bands == ('R','G','B','A'):
        thumb = pil_img.resize((thumbsize,thumbsize))
        SSE, bias = 0, [0,0,0]
        if adjust_color_bias:
            bias = ImageStat.Stat(thumb).mean[:3]
            bias = [b - sum(bias)/3 for b in bias]
        for pixel in thumb.getdata(): # pixel = (R,G,B)
            mu = sum(pixel)/3
            SSE += sum((pixel[i] - mu - bias[i])**2 for i in [0,1,2])
        MSE = SSE/(thumbsize * thumbsize)
        if debug:
            print(f"MSE = {MSE:.4f}")
        if MSE <= MSEthreshold:
            return True
        else:
            return False
    elif len(bands) == 1:
        if debug:
            print(f"Only 1 band (black and white pic).")
        return True
    else:
        raise TypeError(f"PIL bands ({bands}) is not supported.")


def normalize(imgobj, means=None, stds=None, container='np'):
    
    if not means and not stds:
        means, stds = get_mean_std_norms(imgobj)
    elif not means and stds:
        means, _ = get_mean_std_norms(imgobj)
    elif means and not stds:
        _, stds = get_mean_std_norms(imgobj)
    assert len(means) == 3 and len(stds) == 3

    if container == 'np':
        img = to_np(imgobj)


def unnormalize(imgobj, means=None, stds=None, container='np'):
    pass





### ======================================================================== ###
### * ### * ### * ### *        Module API Helpers        * ### * ### * ### * ###
### ======================================================================== ###


def _read_img(filepath, container='np'):
    r"""
    Helper function that coverts image path to [R G B] data (even if gray/bin).
    Does not handle color or size conversions.
        (This is called by to_np/pil/tensor when a path to 1 file is given.)

    Parameters
        filepath - full (or relative) path to image file
        container - output type (can be: np, pil, tensor)
    """
    assert files.is_image(filepath, checkfile=True), \
           f"File ({filepath}) is not a valid img."
    
    pil_img = Image.open(filepath)
    import IPython; IPython.embed(); 
    if pil_img.getbands()[0] == 'P':
        # print('P detected: ', filepath)
        if is_grayscale(pil_img):
            pil_img = pil_img.convert('L')
        else:
            pil_img = pil_img.convert('RGB')
        # print(pil_img.mode, pil_img.getbands(), np.array(pil_img).shape)
    img_bands = pil_img.getbands()

    img = None  # to be set by container conditions
    container = container.lower()
    if container == 'np':
        img = np.uint8(np.array(pil_img))
        # img = cv2.imread(filepath)[...,::-1]  # (H,W,bgr) -> (H,W,rgb)
        # assert img is not None, f"Error reading image ({filepath})"
        # if str(img.dtype) != 'uint8':
        #     img = np.uint8(img)
    elif container == 'pil':
        img = pil_img
    elif 'ten' in container:
        # if files.get_ext(filepath) == '.gif':  # ~3-4x slower than cv2
        #     npimg = np.array(Image.open(filepath).convert('RGB'))  
        # else:
        #     # still ~8-10x faster to copy neg strides than transform
        #     npimg = cv2.imread(filepath)[...,::-1].copy() # (H,W,bgr) -> (H,W,rgb)
        npimg = np.array(pil_img)
        if len(img_bands) > 1:  # >10x faster than PIL transform
            img = torch.from_numpy(np.rollaxis(npimg, 2, 0))
        else:
            img = torch.from_numpy(npimg)
        img = img.float()/255.
    else:
        raise TypeError(f"Container type given ({container}) is not valid.")

    return img


def _resize(imgobj, size=None):
    r"""
    Helper function that resizes single np, pil, or tens image

    Parameters:
        size - (H,W)
    """
    if not size:
        return imgobj
    assert len(size) == 2
    
    if isinstance(imgobj, np.ndarray):  # standardizes array
        return resize(imgobj, (size[0], size[1]))
    elif 'PIL.' in str(type(imgobj)):  
        return imgobj.resize(size[::-1])
    elif isinstance(imgobj, torch.Tensor):
        # if size[0] == size[1]:
        #     return F.interpolate(imgobj, size=size[0])
        # else:
        resized_pil = transforms.ToPILImage()(imgobj).resize(size[::-1])
        return transforms.ToTensor()(resized_pil)
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")
    

def _recolor(imgobj, color=None, extra_channel=False):
    r"""
    Helper function that recolors single np, pil, or tens image

    Parameters:
        color - auto, rgb, gray
        extra_channel - for gray  e.g. np: (H,W) without and (H,W,1) with
    """
    if not color or color == 'auto':
        return imgobj
    
    if isinstance(imgobj, np.ndarray):
        # print('_recolor: ', imgobj.shape)
        assert imgobj.ndim == 3 or imgobj.ndim == 2
        if color == 'rgb':
            if imgobj.ndim == 3 and imgobj.shape[-1] == 3:
                return imgobj
            elif imgobj.ndim == 3 and imgobj.shape[-1] == 1:
                return np.repeat(imgobj, 3, axis=2)
            else:
                return np.repeat(imgobj[:,:,np.newaxis], 3, axis=2)
        else:  # convert to gray
            if imgobj.ndim == 3 and imgobj.shape[-1] == 1:
                if extra_channel:
                    return imgobj
                return imgobj[:,:,0]
            elif imgobj.ndim == 2:
                if extra_channel:
                    return np.repeat(imgobj[:,:,np.newaxis], 1, axis=2)
                return imgobj
            else:  # 3 channel rgb
                # return np.dot(imgobj[...,:3], [0.299, 0.587, 0.114])
                imgobj = np.dot(imgobj, [0.299, 0.587, 0.114])
                if extra_channel:
                    return np.repeat(imgobj[:,:,np.newaxis], 1, axis=2)
                return imgobj
    elif 'PIL.' in str(type(imgobj)):  
        return imgobj.convert(COLOR_2_PILMODE[color])
    elif isinstance(imgobj, torch.Tensor):
        assert imgobj.ndim == 3 or imgobj.ndim == 2
        if color == 'rgb':
            if imgobj.ndim == 3 and imgobj.shape[0] == 3:  # 3xHxW
                return imgobj
            else:  # 1xHxW or HxW
                return imgobj.repeat(3,1,1)
        else:  # convert to gray
            if imgobj.ndim == 3 and imgobj.shape[0] == 1:
                if extra_channel:
                    return imgobj
                return imgobj[0,:,:]
            elif imgobj.ndim == 2:
                if extra_channel:
                    return imgobj.repeat(1,1,1)
                return imgobj
            else:  # 3 channel rgb
                # return np.dot(imgobj[...,:3], [0.299, 0.587, 0.114])
                imgobj = .299*imgobj[0,:,:] + .587*imgobj[1,:,:] + .114*imgobj[2,:,:]
                if extra_channel:
                    return imgobj.repeat(1,1,1)
                return imgobj
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")


def _create_img_list(imgobj):
    r"""
    Returns list of single images in file/path, 3d/4d np/tensor, pil
    """

    imgs = []
    if isinstance(imgobj, str):
        if os.path.isdir(imgobj):
            imgs = files.list_images(imgobj, fullpath=True)
        elif os.path.isfile(imgobj) and files.is_image(imgobj):
            imgs = [imgobj]
        else:
            print(f"WARNING: Path ({imgobj}) is not valid.")
    elif 'PIL.' in str(type(imgobj)):
        imgs = [imgobj]
    elif isinstance(imgobj, np.ndarray) or isinstance(imgobj, torch.Tensor):
        assert imgobj.ndim >= 2 and imgobj.ndim <= 4
        if imgobj.ndim == 4:
            for c in range(imgobj.shape[0]):
                imgs.append(imgobj[c,:,:,:])
        else:
            imgs = [imgobj]
    elif isinstance(imgobj, pd.DataFrame):
        assert 'path' in imgobj.columns
        for path in imgobj['path']:
            imgs.append(path)
    else:
        raise TypeError(f"Image obj of type ({type(imgobj)}) is not valid.")
    
    return imgs



""" Module dmt/utils/parse.py

Parse parameters and display useful error messages.
"""

import string
import numbers



### ------ #      Individual Numbers      # ----- ###

def parse_bool(param, name): 
    """ In python booleans are implemented as ints (so this is in numbers). """
    try:
        param = bool(param)
    except:
        raise ValueError(f'"{name}" must be a boolean or bool convertible.')
    return param


def parse_int(param, name):
    try:
        param = int(param)
    except:
        raise ValueError(f'"{name}" must be an integer or int convertible')
    return param


def parse_nonnegative_int(param, name):
    param = parse_int(param, name)
    
    msg = (f'"{name}" must be 0 or a natural number. '
            f'You gave: {param} (after auto int conversion)')
    assert param >= 0, msg
    
    return param


def parse_positive_int(param, name):
    param = parse_int(param, name)
    
    msg = (f'"{name}" must be a natural number. '
            f'You gave: {param} (after auto int conversion)')
    assert param > 0, msg
    
    return param


def parse_float(param, name):
    try:
        param = float(param)
    except:
        raise ValueError(f'"{name}" must be a float or float-convertible')
    return param


def parse_probability(param, name):
    param = parse_float(param, name)
    
    msg = f'Probability must be a number in [0, 1], not {param}'
    if not (0 <= param <= 1):
        raise ValueError(msg)
    return param


### ------ #      Number Ranges      # ----- ###

def parse_range(nums_range, name, out_min=None, out_max=None, out_type=None):
    r"""
    Args:
        nums_range: single number or tuple of 2 numbers
            If single positive number, -n, n will be returned.
        name: Name of the parameter for an informative error message.
        out_min: Minimal value that range can take,
            default is None, i.e. there is no minimal value.
        out_max: Maximal value that range can take,
            default is None, i.e. there is no maximal value.
        type_constraint: Precise type output range must take.
    Returns:
        A tuple of two numbers (min_range, max_range).
    """
    if isinstance(nums_range, numbers.Number):  # single number given
        if nums_range < 0:
            raise ValueError(
                f'If {name} is a single number,'
                f' it must be positive, not {nums_range}')
        if out_min is not None and nums_range < out_min:
            raise ValueError(
                f'If {name} is a single number, it must be greater'
                f' than {out_min}, not {nums_range}'
            )
        if out_max is not None and nums_range > out_max:
            raise ValueError(
                f'If {name} is a single number, it must be smaller'
                f' than {out_max}, not {nums_range}'
            )
        if out_type is not None:
            if not isinstance(nums_range, out_type):
                raise ValueError(
                    f'If {name} is a single number, it must be of'
                    f' type {out_type}, not {type(nums_range)}'
                )
        return (-nums_range, nums_range)

    try:
        min_value, max_value = nums_range
    except (TypeError, ValueError):
        raise ValueError(
            f'If {name} is not a single number, it must be'
            f' a sequence of len 2, not {nums_range}'
        )

    min_is_number = isinstance(min_value, numbers.Number)
    max_is_number = isinstance(max_value, numbers.Number)
    if not min_is_number or not max_is_number:
        message = (
            f'{name} values must be numbers, not {nums_range}')
        raise ValueError(message)

    if min_value > max_value:
        raise ValueError(
            f'If {name} is a sequence, the second value must be'
            f' equal or greater than the first, but it is {nums_range}')

    if out_min is not None and min_value < out_min:
        raise ValueError(
            f'If {name} is a sequence, the first value must be greater'
            f' than {out_min}, but it is {min_value}'
        )

    if out_max is not None and max_value > out_max:
        raise ValueError(
            f'If {name} is a sequence, the second value must be smaller'
            f' than {out_max}, but it is {max_value}'
        )

    if out_type is not None:
        min_type_ok = isinstance(min_value, out_type)
        max_type_ok = isinstance(max_value, out_type)
        if not min_type_ok or not max_type_ok:
            raise ValueError(
                f'If "{name}" is a sequence, its values must be of'
                f' type "{out_type}", not "{type(nums_range)}"'
            )
    return nums_range


""" timers.py (By: Charley Zhang, July 2020)
Tools to keep track of running-time and print them neatly.
"""


import time
from collections import OrderedDict

class StopWatch:

    def __init__(self):
        self.reset()

    def tic(self, name='default'):
        self.stops[name] = 0
        self.starts[name] = time.time()
        return self

    def toc(self, name='default', disp=True):
        if name not in self.starts:
            print(f"StopWatch({name}) did not get tic'd yet.")
            return
        
        self.stops[name] = time.time()
        if disp:
            self.disp(name=name)
        return self.stops[name] - self.starts[name]

    def disp(self, name='default'):
        if name not in self.stops or name not in self.starts:
            print(f"StopWatch({name}) did not get tic'd yet.")
            return
        
        elapsed = self.stops[name] - self.starts[name]
        if elapsed > 0:
            if elapsed > 4200:
                print(f"StopWatch({name}) took {elapsed/3600:.2f} hrs")
            elif elapsed > 60:
                print(f"StopWatch({name}) took {elapsed/60:.2f} min")
            else:
                print(f"StopWatch({name}) took {elapsed:.2f} sec")
        else:
            print(f"StopWatch({name}) did not get toc'd for its tic.")

    def reset(self):
        self.starts = {}  # time in seconds
        self.stops = {}
        return self

        


""" Module dmt.__init__.py
Top level package for dmt (Deep Medical Toolkit).
	- Deep: for use in deep learning pipelines (training, inference)
	- Medical: medical imaging tasks (2D/3D classification, localization, etc)
	- Tools: common functionality for data storage, loading, and transforms  
"""

__author__ = 'Charley Yejia Zhang'
__email__ = 'yzhang46@nd.edu'
__version__ = '0.0.1'

__all__ = []

from .constants import IMAGE, SHAPE, LOCATION
from . import data
from . import metrics
from . import losses
# from . import nets
# import transforms





import sys
import torch, torch.nn as nn


class CrossEntropy(nn.Module):

    def __init__(self, weights=None):
        super(CrossEntropy, self).__init__()
        self.weights = torch.tensor(weights).to(torch.float32) if weights \
            else None
        self.CE = nn.CrossEntropyLoss(weight=self.weights)
        print(f"\tCE loss initialized (weights:{weights}).")
    
    def forward(self, preds, targs):
        if targs.ndim == 4:
            targs = targs.squeeze()
        assert targs.ndim == 3, f"Invalid target dims: {targs.shape}"
        
        loss = self.CE(preds, targs.long())
        return loss


class FocalLoss(nn.Module):
    """ L(p_t) = -alpha_t * (1-p_t)^gamma * log(p_t)
          p_t: prediction prob for target class, t
    Paper: Focal Loss for Dense Object Detection
        https://arxiv.org/pdf/1708.02002.pdf
    """

    def __init__(self, alpha=1, gamma=2):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, preds, targs, eps=10**-6):
        B = pred.shape[0]
        preds = self.softmax(preds)       
        preds = inputs.view(B, -1)
        targets = targets.view(B, -1)
        
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        BCE_EXP = torch.exp(-BCE)
        focal_loss = self.alpha * (1 - BCE_EXP)**self.gamma * BCE
                       
        return focal_loss


class SoftDiceLoss(nn.Module):

    def __init__(self, square_denom=False):
        super(SoftDiceLoss, self).__init__()
        self.square_denom = square_denom
        self.softmax = nn.Softmax(dim=1)
        print('\tSoftDice initialized.')

    def forward(self, pred, targ, eps=1):
        """
        Parameters
            pred: BxCxHxW logits as a float tensor on the same device
            targ: BxCxHxW binary labels as a float tensor
        """
        assert pred.shape == targ.shape, f"Pred: {pred.shape}, Targ: {targ.shape}"
        assert pred.dtype == torch.float32
        
        B, C = pred.shape[0], pred.shape[1]
        pred_probs = pred.sigmoid() if C == 1 else self.softmax(pred)
        preds = pred_probs.view(B, -1)
        targs = targ.view(B, -1)
        
        intersect = preds * targs
        if self.square_denom:
            dice = (2. * intersect.sum(1) + eps) / (preds.square().sum(1) + \
                    targs.square().sum(1) + eps)
        else:
            dice = (2. * intersect.sum(1) + eps) / (preds.sum(1) + \
                    targs.sum(1) + eps)
        
        return 1 - dice.mean()


class SoftJaccardLoss(nn.Module):

    def __init__(self, log_loss=False):
        """
        log_loss = -ln(IOU) as presented in this paper:
            https://arxiv.org/pdf/1608.01471.pdf
        """
        super(SoftJaccardLoss, self).__init__()
        self.log_loss = log_loss
        self.softmax = nn.Softmax(dim=1)
        print(f'\tSoftJaccard initialized (log_jaccard: {log_loss}).')

    def forward(self, pred, targ, eps=10**-6):
        """
        Parameters
            pred: BxCxHxW logits as a float tensor on the same device
            targ: BxCxHxW binary labels as a float tensor
        """
        assert pred.shape == targ.shape, f"Pred: {pred.shape}, Targ: {targ.shape}"
        assert pred.dtype == torch.float32
        
        B, C = pred.shape[0], pred.shape[1]
        pred_probs = pred.sigmoid() if C == 1 else self.softmax(pred)
        preds = pred_probs.view(B, -1)
        targs = targ.view(B, -1)
        
        intersect = torch.sum(preds * targs, 1)
        union = torch.sum(preds + targs, 1) - intersect  # B x C tensor
        iou = torch.mean((intersect + eps) / (union + eps))
        return 1 - iou if not self.log_loss else -torch.log(iou)


class FocalTverskyLoss(nn.Module):
    """ Generalized loss that is a superset of dice, jaccard, and focal loss."""

    def __init__(self, alpha=0.5, beta=0.5, gamma=1):
        """ L = (1 - TI)^gamma, TI = TP / (TP + alpha*FN + beta*FP) 
        Quick Notes
            - alpha = beta = 0.5 is dice loss
            - alpha = beta = 1. is jaccard loss
            - gamma > 1 gives concave loss, gamma < 1 gives convex
            - good default for FN-focus: alpha=0.7, beta=0.3, gamma=3/4
        """
        super(FocalTverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        print(f"\tTversky loss initialized (α:{alpha}, β:{beta}, γ:{gamma}).")

    def forward(self, pred, targ, eps=1):
        assert pred.shape == targ.shape, f"Pred{pred.shape}, Targ{targ.shape}"
        assert pred.ndim == 4, f"4D tensor required. Got: {pred.shape}."
        
        B, C = pred.shape[0], pred.shape[1]
        if C == 1:
            pred_probs = pred.sigmoid()
        else:
            smax = torch.nn.Softmax(dim=1)
            pred_probs = smax(pred)
        preds = pred_probs.view(B, -1)
        targets = targ.view(B, -1)
        
        TP = (preds * targets).sum(1)
        FN = (targets * (1 - preds)).sum(1)
        FP = ((1 - targets) * preds).sum(1)
        
        tversky = (TP + eps) / (TP + self.alpha * FN + self.beta * FP + eps)
        return (1 - torch.mean(tversky)) ** self.gamma




import psutil


def is_process_running(pid_name):
    '''
    Check if there is any running process that contains the given name processName.
    '''
    
    if isinstance(pid_name, int):
        return is_process_running_from_pid(pid_name)
    elif isinstance(pid_name, str):
        for proc in psutil.process_iter():
            try:
                # Check if process name contains the given name string.
                if pid_name.lower() in proc.name().lower():
                    return True
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                pass
    return False

def is_process_running_from_pid(pid):
    for proc in psutil.process_iter():
        if proc.pid == pid:
            return True
    return False


""" Module __init__.py for dmt.losses """


### ------ #     Classification Losses     # ----- ###

### ------ #     Localization Losses     # ----- ###

### ------ #     Segmentation Losses     # ----- ###




""" Module ramps.py
Functions for ramping hyperparameters up or down
Each function takes the current training step or epoch, and the
ramp length in the same format, and returns a multiplier between
0 and 1.

Taken from PyTorch Mean-Teacher repo:
https://github.com/CuriousAI/mean-teacher/blob/master/pytorch/mean_teacher/ramps.py
"""

import numpy as np


def sigmoid_rampup(current, rampup_length):
    """Exponential rampup from https://arxiv.org/abs/1610.02242"""
    if rampup_length == 0:
        return 1.0
    else:
        current = np.clip(current, 0.0, rampup_length)
        phase = 1.0 - current / rampup_length
        return float(np.exp(-5.0 * phase * phase))


def linear_rampup(current, rampup_length):
    """Linear rampup"""
    assert current >= 0 and rampup_length >= 0
    if current >= rampup_length:
        return 1.0
    else:
        return current / rampup_length


def cosine_rampdown(current, rampdown_length):
    """Cosine rampdown from https://arxiv.org/abs/1608.03983"""
    assert 0 <= current <= rampdown_length
    return float(.5 * (np.cos(np.pi * current / rampdown_length) + 1))


""" Module utils/io/images3d.py 
Contains common utilities for 3D images (most functionality through sitk).
"""

import sys, os
import SimpleITK as sitk
import numpy as np

from . import files as file_utils

### Constants ###
IMAGE_EXTS = ['.bmp', '.png', '.jpg', '.jpeg', '.tif', '.tiff', '.gif',
              '.nii', '.nii.gz', '.dcm']


### ======================================================================== ###
### * ### * ### * ### *          API Definitions         * ### * ### * ### * ### 
### ======================================================================== ###

__all__ = ['read_sitk3d',
           'write_sitk_gray3d', 'write_np_gray3d',
           'to_np_channel_last', 'to_np_channel_first'
          ]

### ---- ### ---- \\    Image Read/Write     // ---- ### ---- ###

def read_sitk3d(im_path, pixtype=sitk.sitkInt16):
    image = sitk.ReadImage(im_path, pixtype)
    dims = image.GetNumberOfComponentsPerPixel()
    assert dims == 1, f"Only 3D gray sitk images are supported."
    return image


def write_sitk_gray3d(sitk_im, path, compress=False):
    r""" Save 3D sitk grayscale image. 
    Args:
        sitk_im: SimpleITK.Image
        path: str or pathlib.Path
        compress: bool
            If save-type is .nii or .nii.gz, doesn't affect output.
    """
    channels = sitk_im.GetNumberOfComponentsPerPixel()
    assert channels == 1, f"Only 3D gray sitk images are supported."
    _write_sitk_image(sitk_im, path, compress=compress)


def write_np_gray3d(np_im, path, extra_channel=False, compress=False):
    r""" Save 3D grayscale image of shape 1xHxWxD, HxWxD, or HxWxDx1 """
    if extra_channel:
        np_im = np_im.squeeze(0).squeeze(-1)
    assert np_im.ndim == 3, f"Only 3D gray images are supported."
    img = sitk.GetImageFromArray(np_im, isVector=False, compress=False)
    _write_sitk_image(img, path, compress=compress)


def _write_sitk_image(img, path, compress=False):
    r""" Creates directory structure and writes using sitk's writer. 
    Arguments:
        img (sitk image)
        path (str)
    Returns:
        True / False (bool) if successful or failed.
    """ 
    file_utils.create_dirs_from_file(path)
    try:
        writer = sitk.ImageFileWriter()
        writer.SetFileName(path)
        writer.SetUseCompression(compress)
        writer.Execute(img)
        return True
    except:
        print(f"Error writing image to {path}!")
        file_utils.delete_file(path)
        return False


### ---- ### ---- \\    Common np Operations     // ---- ### ---- ###


def hu_to_np_image(sitk_np_im, clip=[-1024, 325], scale=[0, 255], 
                  outtype=np.uint8, rgb=False):
    r""" 1. Clip  2. Normalize + Scale  3. Cast Type (opt to RGB) """
    np_im = sitk_np_im
    if isinstance(sitk_np_im, sitk.Image):
        np_im = sitk.GetImageFromArray(sitk_np_im, isVector=False)
        np_im = np_im.astype(np.float32)
    assert isinstance(np_im, np.ndarray)
    if clip:
        np_im = np.clip(np_im, clip[0], clip[1])
    
    # normalize between 0 & 1
    im_min, im_max = np_im.min(), np_im.max()
    np_im = (np_im - im_min) / (im_max - im_min + 1e-6)
    if scale:
        assert scale[1] > scale[0]
        np_im = np_im * (scale[1] - scale[0]) - scale[0]
    if rgb:
        np_im = np.stack([np_im, np_im, np_im], axis=-1)
    if outtype is np.uint8:
        np_im = np.round(np_im)    
    np_im = np_im.astype(outtype)
    return np_im


### ---- ### ---- \\    Common sitk Operations     // ---- ### ---- ###


def resample_sitk_isotropic(sitk_im, spacing=None, smooth=0,
                             interpolation='nearest'):
    r"""
    Args:
        sitk_im: sitk.Image
            sitk image object to be made isotropic.
        ?spacing: None or float
            If None, then uses minimum spacing as reference, else the given one.
        ?smooth: None or False or positive float
            Applies Gaussian smoothing before resampling.
        ?interpolation: str ∈ {'nearest', 'linear'}
            Apply either linear or nearest neighbor interpolation
    """
    orig_size = sitk_im.GetSize()
    ndim = len(orig_size)
    orig_spacing = sitk_im.GetSpacing()
    if all(s == orig_spacing[0] for s in orig_spacing):
        return sitk_im
    
    ref_spacing = min(orig_spacing) if spacing is None else spacing
    new_spacing = [ref_spacing] * ndim
    new_size = [int(round(orig_size[i] * orig_spacing[i] / ref_spacing)) \
                for i in range(ndim)]
    if 'linear' in interpolation:
        interpolation = sitk.sitkLinear
    elif 'spline' in interpolation or 'cubic' in interpolation:
        interpolation = sitk.sitkBSpline
    else:
        interpolation = sitk.sitkNearestNeighbor
    
    if smooth and smooth > 0:
        sitk_im = sitk.SmoothingRecursiveGaussian(sitk_im, smooth)
    direction = sitk_im.GetDirection()
    resampled_im = sitk.Resample(sitk_im, new_size, sitk.Transform(),
                                 interpolation, sitk_im.GetOrigin(),
                                 new_spacing, direction, 0, 
                                 sitk_im.GetPixelIDValue())
    return resampled_im


### ---- ### ---- \\    Image Transforms     // ---- ### ---- ###


### Image Storage Transforms

def sitk_to_np(sitk_im):
    return sitk.GetArrayFromImage(sitk_im)


def np_to_sitk(np_im, has_channels=False):
    r"""
    Args:
        has_channels: bool
            If image is rgb, has_channels=True
    """
    return sitk.GetImageFromArray(np_im, isVector=has_channels)
    

def to_np_channel_last(np_im):
    if np_im.shape[0] not in range(1, 6):
        print(f"(to_np_channel_last) WARNING: Weird image {np_im.shape}")
    return np.moveaxis(np_im, 0, -1)


def to_np_channel_first(np_im):
    if np_im.shape[-1] not in range(1, 6):
        print(f"(to_np_channel_first) WARNING: Weird image {np_im.shape}")
    return np.moveaxis(np_im, -1, 0)
        



import torch

from .sample import Sample


class SampleSet(torch.utils.data.Dataset):
    """ Base container object for a dataset's subset (e.g. KiTS train).
    
    Intended Functionality:
      (1) Serve as a container for a data subset (training, val set)
      (2) Return a multi-processing iterator for preprocessed images.
           For 3D images, this could be resampling, for 2D, it could just
           pre-load the data in memory using a Dataloader wrapper. 
    """
    
    def __init__(
            self, 
            samples,
            transforms=None,  # intended for preprocessing of entire images
            ):
        self.transforms = transforms
        samples = self._parse_samples(samples)
        self._samples = samples
        
    @property
    def samples(self):
        return self._samples
        
    def __len__(self):
        return len(self.samples)
    
    def __getitem__(self, idx):
        if idx < 0 or idx > len(self):
            raise ValueError(f"Index {idx} is out of bounds.")
        sample = self.samples[idx]
        if self.transforms is not None:
            sample = self.transforms(sample)
        return sample
        
    def _parse_samples(self, samples):
        try:
            iter(samples)
        except TypeError as e:
            msg = f'Samples must be an iterable, not {type(samples)}'
            raise TypeError(msg) from e
        
        if not samples:
            raise ValueError('Given samples list is empty')

        for sample in samples:
            if not isinstance(sample, Sample):
                msg = (
                    'Samples list must contain instances of sample.Sample,'
                    f' not "{type(sample)}"'
                )
                raise TypeError(msg)
        return samples
    
    def __repr__(self):
        string = (f'SampleSet Object: {len(self)} samples.')
        return string





import colorsys
import scipy
import matplotlib.pyplot as plt
import numpy as np
import SimpleITK as sitk

from ..io import images2d as i2d_utils, images3d as i3d_utils
from . import np_image as npi_utils



### ---- ### ---- \\    3D & 2D Plotting Functions     // ---- ### ---- ### 


def plot_3d_slices(np_volume, mask=None,
                   title='', slices=15, image_size=5, save=''):
    """ Plots sagittal, coronal, and axial slices at centered intervals of img.
    """
    assert np_volume.ndim == 3, f"Only 3D grayscale images HxWxD accepted."
    shape = np_volume.shape
    if mask is not None:
        assert shape == mask.shape, f"Mask and image shapes should match."
        mask_rgb = mask_to_color(mask)
    
    matrix_of_slices, matrix_of_titles = [], []
    for dim in range(np_volume.ndim):
        dim_slices, dim_titles = [], []
        dim_range = range(0, np_volume.shape[dim], np_volume.shape[dim] // slices)
        for dim_index in dim_range:
            if dim == 0:  # axial
                if mask is not None:
                    mslice = mask_rgb[dim_index, ...]
                slice = np_volume[dim_index, :, :]
                title = 'axial: ' + _get_plot_title(slice)
            elif dim == 1:
                if mask is not None:
                    mslice = mask_rgb[:, dim_index, ...]
                slice = np_volume[:, dim_index, :]
                title = 'sagittal: ' + _get_plot_title(slice)
            else:
                if mask is not None:
                    mslice = mask_rgb[:, :, dim_index, ...]
                slice = np_volume[:, :, dim_index]
                title = 'axial: ' + _get_plot_title(slice)
            slice = i3d_utils.hu_to_np_image(slice, clip=[-1024, 325], 
                                             scale=[0, 255], rgb=True)
            if mask is not None:
                slice = np.round(0.3 * mslice + 0.7 * slice).astype(np.uint8)
            dim_slices.append(slice)
            dim_titles.append(title)
        matrix_of_slices.append(dim_slices)
        matrix_of_titles.append(dim_titles)
    npi_utils.print_image_summary(np_volume)
    plot_grid(matrix_of_slices, titles=matrix_of_titles, image_size=image_size,
              save=save)


def plot_2d_image(np_image, title='', image_size=10, save=''):
    assert npi_utils.is_standard_2dimage(np_image), 'Image not standard!'
    axis_title = _get_plot_title(np_image)
    np_image = npi_utils.standardize_2dimage(np_image)
    titles = [[axis_title]] if not title else [[axis_title + ' | ' + title]]
    plot_grid([[np_image]], titles=titles, image_size=image_size, save=save)
    

def plot_2d_images(np_images_list, titles=[], cols=3, image_size=10, save=''):
    if len(np_images_list) == 0:
        return
    if titles:
        assert len(np_images_list) == len(titles)
    
    R = (len(np_images_list) // cols) + 1
    matrix_of_images, matrix_of_titles = [], []
    for r in range(R):
        images_row, titles_row = [], []
        for c in range(cols):
            curr_idx = r * cols + c
            if curr_idx >= len(np_images_list):
                images_row.append(None)
                titles_row.append('')
                continue
            np_image = np_images_list[curr_idx]
            axis_title = _get_plot_title(np_image)
            custom_title = '' if not titles else ' | ' + titles[curr_idx]
            
            assert npi_utils.is_standard_2dimage(np_image), 'Image not standard!'
            
            np_image = npi_utils.standardize_2dimage(np_image)
            axis_title = axis_title + custom_title
            images_row.append(np_image)
            titles_row.append(axis_title)
        matrix_of_images.append(images_row)
        matrix_of_titles.append(titles_row)
    plot_grid(matrix_of_images, titles=matrix_of_titles, 
              image_size=image_size, save=save)


def plot_grid(matrix_of_images, titles=None, image_size=10, save=''):
    R = len(matrix_of_images)
    max_C = max([len(a) for a in matrix_of_images])
    fig = plt.figure(figsize=(max_C * image_size, R * image_size))
    for r in range(R):
        C = len(matrix_of_images[r])
        for c in range(C):
            position = r * C + (c + 1)
            # print(position)
            # print(titles[r][c])
            title = titles[r][c] if titles is not None else ''
            image = matrix_of_images[r][c]
            if image is not None:
                ax = fig.add_subplot(R, C, position)
                ax.set_title(title)
                ax.imshow(image, cmap='gray' if image.ndim == 2 else None)
    # plt.subplots_adjust(left=1, bottom=0, right=2, top=1, wspace=0, hspace=0)
    plt.show()
    if save:
        plt.savefig(save)
    
    
def _get_plot_title(np_image):
    shape = np_image.shape
    shape_str = 'x'.join([str(s) for s in shape])
    max_val = np_image.max()
    max_str = f"{max_val}" if 'int' in str(type(max_val)) else f"{max_val:.2f}"
    min_val = np_image.min()
    min_str = f"{min_val}" if 'int' in str(type(min_val)) else f"{min_val:.2f}"
    return f"{shape_str} max:{max_str} min:{min_str}"


### ---- ### ---- \\    Segmentation Mask Functions     // ---- ### ---- ###


# def mask_to_color(np_mask, channel_last=True):
#     """
#     Arguments:
#         np_mask - CxHxW(xD) 
#     """
#     if 'bool' not in np_mask.dtype:
#         np_mask = np_mask > 0
#     class_channel = -1 if channel_last else 0
#     num_classes = np_mask.shape[class_channel]
#     rgb_colors = generate_colors(num_classes)
#     new_mask = np.zeros(tuple(list(np_mask.shape) + [3]))
#     for cnum in range(1, num_classes):  # cnum 0 = background
#         new_mask[np_mask == cnum] = rgb_colors[cnum]
#     return new_mask


def mask_to_color(sitk_np_im, has_channels=False):
    if has_channels:  # each class label defined in its own channel.
        raise NotImplementedError()
    np_im = sitk_np_im
    if isinstance(sitk_np_im, sitk.Image):
        np_im = sitk.GetImageFromArray(sitk_np_im)
    assert isinstance(np_im, np.ndarray)
    class_values = np.unique(np_im)
    colors = generate_colors(len(class_values))
    
    colored_mask = np.zeros(list(np_im.shape) + [3]).astype(np.uint8)
    for c in range(len(class_values)):
        colored_mask[np_im == class_values[c]] = colors[c]
    return colored_mask


def overlay_mask(volume_ims, segmentation_ims, segmentation, alpha):
    # Get binary array for places where an ROI lives
    segbin = np.greater(segmentation, 0)
    repeated_segbin = np.stack((segbin, segbin, segbin), axis=-1)
    # Weighted sum where there's a value to overlay
    overlayed = np.where(
        repeated_segbin,
        np.round(alpha*segmentation_ims+(1-alpha)*volume_ims).astype(np.uint8),
        np.round(volume_ims).astype(np.uint8)
    )
    return overlayed


def generate_colors(num_colors, black_first=True):
    if black_first:
        HSVs = [[0, 0, 0]]
        if num_colors > 1:
            HSVs += [[x / (num_colors - 1), 0.5, 0.5] for x in range(num_colors - 1)]
    else:
        HSVs = [[x / num_colors, 0.5, 0.5] for x in range(num_colors)]
    RGBs = list(map(lambda x: colorsys.hsv_to_rgb(*x), HSVs))
    RGBs = [[int(255 * x) for x in rgb] for rgb in RGBs]
    return RGBs



import logging
import weakref
import pprint

import dmt
from ..image_base import Image
from ..label_base import Label


class Sample(dict):
    """ Base class for all data samples (like subjects in TorchIO).
    
    This object contains a counter to be used for global sample identification.
    """
    _id_counter = 0
    _gid_to_instance = weakref.WeakValueDictionary()
    _reserved_attributes = ('gid', 'num_images', 'num_labels')
    
    def __init__(self, dictionary=None, **kwargs):
        """
        Args:
            dictionary: dict of attributes to be updated in sample
            excluded_images: 
        """
        # Global Sample ID tracking
        self.gid = Sample._id_counter  # global ID for Sample objects
        Sample._gid_to_instance[self.gid] = self
        Sample._id_counter += 1
        
        self.applied_transforms = []
        if dictionary:
            assert isinstance(dictionary, dict), "The type is in the name man.."
            kwargs.update(dictionary)
        super().__init__(**kwargs)
        self.update_attributes()
        logging.debug(f'📦 Sample object created: {repr(self)}')
    
    @property
    def num_images(self):
        """ Returns number of non-label images in this sample. """
        return len(self.get_image_keys(include_labels=False))
    
    @property
    def num_labels(self):
        """ Returns number of image labels (doesn't include categories). """
        return len(self.get_label_keys(only_images=True))
    
    
    ### ------ #      Functionality for Image Retrieval      # ----- ###
    
    def get_image_keys(self, include_labels=False):
        """ Returns a list of key names that are image objects. """
        image_keys = []
        for k, v in self.items():
            if isinstance(v, dmt.data.Image):
                if not include_labels and isinstance(v, dmt.data.Label):
                    continue
                image_keys.append(k)
        return image_keys
    
    def get_images_dict(self, include_labels=False):
        image_keys = self.get_image_keys(include_labels=include_labels)
        images_dict = {k: self[k] for k in image_keys}
        return images_dict
    
    def get_label_keys(self, only_images=True):
        """ Returns a list of key names that are labels. """
        label_keys = []
        for k, v in self.items():
            if isinstance(v, dmt.data.Label):
                if only_images and not isinstance(v, Image):
                    continue
                label_keys.append(k)
        return label_keys
    
    def get_labels_dict(self, only_images=True):
        label_keys = self.get_label_keys(only_images=only_images)
        labels_dict = {k: self[k] for k in label_keys}
        return labels_dict
    
    def get_target_image_keys(self, include_keys=None, exclude_keys=None):
        include_keys, exclude_keys = Sample._parse_include_exclude_keys(
            include_keys, exclude_keys)
        candidate_image_keys = set(self.get_image_keys(include_labels=True))
        if include_keys:
            candidate_image_keys = candidate_image_keys.intersection(
                                    include_keys)
        for k in exclude_keys:
            if k in candidate_image_keys:
                candidate_image_keys.remove(k)
        return list(candidate_image_keys)
    
    @staticmethod
    def _parse_include_exclude_keys(include_keys, exclude_keys):
        if exclude_keys is None:
            exclude_keys = []
        elif isinstance(exclude_keys, str):
            exclude_keys = [exclude_keys]
        msg = ('Keys in "exclude_keys" must be a string or seqence of strings. '
               f'Got: {exclude_keys}')
        for key in exclude_keys:
            assert isinstance(key, str), msg
        
        new_include_keys = []
        if include_keys is None:
            include_keys = []
        elif isinstance(include_keys, str):
            include_keys = [include_keys]
        msg = ('Keys in "include_keys" must be a string or seqence of strings. '
               f'Got: {include_keys}')
        for key in include_keys:
            assert isinstance(key, str), msg
            if key in exclude_keys:
                continue
            new_include_keys.append(key)
        
        return new_include_keys, exclude_keys
    
    
    ### ------ #      Transforms History & Reproducibility      # ----- ###
    
    def record_transform(self, transform_reproducing_arguments):
        item = transform_reproducing_arguments
        self.applied_transforms.append(item)
        logging.debug(f'Sample {self.gid}: Updating transforms {item}. \n'
                      f'  Transforms history: {self.applied_transforms}.')
    
    
    ### ------ #      Other Functionality      # ----- ###
    
    def __repr__(self):
        string = (
            f'{self.__class__.__name__} (gid={self.gid}) \n'
            f'{self.num_images} Image(s) \n'
        )
        for name, img in self.get_images_dict().items():
            perm = 'RAM-loaded' if img.permanent_load else 'lazy-load'
            t = img.type
            string += f'  "{name}": {img.__class__.__name__} ({perm}, {t})\n'
            if img.permanent_load:
                string += f'   {img.shape}, mem={img.memory/1e6} MB \n'
            else:
                string += f'   "{img.path}"\n'
        string += f'{self.num_labels} Label(s) \n'
        for name, lab in self.get_labels_dict().items():
            string += f'  "{name}": {lab.__class__.__name__} \n'
            if not hasattr(lab, 'permanent_load') or lab.permanent_load:
                string += f'   Has ids: {lab.unique_values} \n'
        
        hidden_keys = list(self.get_images_dict().keys()) + \
           list(self.get_labels_dict().keys()) + list(self._reserved_attributes)
        attrs = {k: v for k, v in self.items() if k not in hidden_keys
                                                and k[0] != '_'}
        attrs_str = pprint.pformat(attrs, indent=2, compact=True)
        string += f"Other Attributes: \n {attrs_str} \n"
        return string
                
        return string
    
    def __len__(self):
        """ Returns the number of image & label items. """
        return len(self.get_images_dict()) + len(self.get_labels_dict)
    
    def __copy__(self):
        print('Sample copy called!')
        import copy
        result_dict = {}
        for key, value in self.items():
            if isinstance(value, Image):
                value = copy.copy(value)
            else:
                value = copy.deepcopy(value)
            result_dict[key] = value
        new = Sample(result_dict)
        new.applied_transforms = self.applied_transforms[:]  # shallow copy
        return new
    
    ### Functions below offers compatible/correct dict functionality
    ###   i.e. must update __dict__ after every update & override default funcs
    
    def __setattr__(self, name, value):
        """ Setting attrs in dot notation requires updating dict. """
        name = self._parse_new_attribute(name)
        self[name] = value
        super().__setattr__(name, value)
    
    def __getitem__(self, item):
        item = super().__getitem__(item)
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {item}')
        return item
    
    def __setitem__(self, key, value):
        key = self._parse_new_attribute(key)
        super().__setitem__(key, value)
        self.update_attributes()
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {value}')
        
    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                msg = f"Update expected at most 1 arguments, got {len(args)}."
                raise TypeError(msg)
            for k, v in dict(args[0]).items():
                k = self._parse_new_attribute(k)
                self[k] = v
        for k, v in kwargs.items():
            k = self._parse_new_attribute(k)
            self[k] = v
        self.update_attributes()
            
    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        self.update_attributes()
        return self[key]
    
    def update_attributes(self):
        # Allows attribute access through dot notation, e.g. image.spacing
        self.__dict__.update(self)
        
    def _parse_new_attribute(self, attribute):
        return attribute
        
        # TODO gatekeep user attribute setting
        msg = (f'Given attribute "{attribute}" is a reserved attribute '
               f'for class {type(self).__name__}.')
        assert attribute not in self._reserved_attributes, msg
        
        msg = f'Given attribute "{attribute}" cannot start with "_".'
        assert attribute[0] != '_', msg
        
        return attribute
        
    

        


""" denseunet.py (modified version of 2D DenseUNet by: Charley Zhang)
Modified for readability + added modular functionality for densenet
backbones other than 161.

Original DenseUNet 161 implementation:
https://github.com/xmengli999/TCSM/blob/master/models/network.py

DenseUNet-121: 13.1M params
DenseUNet-169: 23.7M params
DenseUNet-201: 31.8M params
DenseUNet-161: 49.5M params
"""


from torch import nn
import torch.nn.functional as F
import torch

from .. import densenet 
from ..base_model import BaseModel


densenet_outdims = {
    # [7x7+MP, denseblock1, denseblock2, denseblock3, denseblock4 output dims]
    'densenet121': [64, 256, 512, 1024, 1024],
    'densenet169': [64, 256, 512, 1280, 1664],
    'densenet201': [64, 256, 512, 1792, 1920],
    'densenet161': [96, 384, 768, 2112, 2208]
}


def get_model(form='densenet121', pretrained=True, 
              out_channels=1, deconv=False):
    """
    Args:
        name: Name of the DenseNet encoder to use.
        pretrained: ImageNet pretrained encoder.
        out_channels: Number of output channels.
        deconv: Flag to use deconv or interpolation when upsampling.
    """
    model = DenseUNet2D(
                str(form), 
                pretrained=pretrained, 
                out_channels=out_channels, 
                deconv=deconv
                )
    return model


### ======================================================================== ###
### * ### * ### * ### *          Model Components        * ### * ### * ### * ###
### ======================================================================== ###


class UpBlock(nn.Module):
    def __init__(self, side_in_dim, bot_in_dim, out_dim, deconv=False):
        super(UpBlock, self).__init__()
        
        self.dim_mismatched = side_in_dim != bot_in_dim
        if self.dim_mismatched:
            self.side_conv = nn.Conv2d(side_in_dim, bot_in_dim, 1, padding=0)
            nn.init.xavier_normal_(self.side_conv.weight)
        
        self.aggregated_conv = nn.Conv2d(bot_in_dim, out_dim, 3, padding=1)
        nn.init.xavier_normal_(self.aggregated_conv.weight)

        self.bn = nn.BatchNorm2d(out_dim)
        nn.init.constant_(self.bn.weight, 1)
        nn.init.constant_(self.bn.bias, 0)

        self.use_deconv = deconv
        if self.use_deconv:
            self.deconv = nn.ConvTranspose2d(bot_in_dim, bot_in_dim, 3, 
                stride=2, padding=1, output_padding=1)
            nn.init.xavier_normal_(self.deconv.weight)
        
    def forward(self, side_in, up_in):
        if self.use_deconv:
            up_in = self.deconv(up_in)
        else:
            up_in = F.interpolate(up_in, scale_factor=2, mode='bilinear', 
                align_corners=True)
        
        if self.dim_mismatched:
            side_in = self.side_conv(side_in)
        
        agg = torch.add(up_in, side_in)   # no cat like U-Net
        out = F.relu(self.bn(self.aggregated_conv(agg)))

        return out


class ForwardHook():
    features = None
    
    def __init__(self, module): 
        self.hook = module.register_forward_hook(self.hook_fn)

    def hook_fn(self, module, input, output): 
        self.features = output

    def remove(self): 
        self.hook.remove()


### ======================================================================== ###
### * ### * ### * ### *             Main Entree          * ### * ### * ### * ###
### ======================================================================== ###


class DenseUNet2D(BaseModel):

    def __init__(self, name='densenet161', out_channels=1, pretrained=True, 
            deconv=False):
        super(DenseUNet2D, self).__init__()
        
        self.features, self.encoder_hooks, ldims = self._setup_encoder(
            name, 
            pretrained=pretrained
        )
        self.classifier = nn.Linear(1024, 3)  # HARD CODED
        
        self.up1 = UpBlock(ldims[3], ldims[4], ldims[2], deconv=deconv)
        self.up2 = UpBlock(ldims[2], ldims[2], ldims[1], deconv=deconv)  
        self.up3 = UpBlock(ldims[1], ldims[1], ldims[0], deconv=deconv)
        self.up4 = UpBlock(ldims[0], ldims[0], ldims[0], deconv=deconv)

        self.conv1 = nn.Conv2d(ldims[0], 64, kernel_size=3, padding=1)
        self.bn1 = nn.BatchNorm2d(64)
        self.conv2 = nn.Conv2d(64, out_channels, kernel_size=1, padding=0)
        nn.init.xavier_normal_(self.conv1.weight)
        nn.init.xavier_normal_(self.conv2.weight)
        nn.init.constant_(self.bn1.weight, 1)
        nn.init.constant_(self.bn1.bias, 0)

        self.use_deconv = deconv
        if self.use_deconv:
            self.final_deconv = nn.ConvTranspose2d(ldims[0], ldims[0], 3, 
                stride=2, padding=1, output_padding=1)
            nn.init.xavier_normal_(self.final_deconv.weight)

        print(f"\tInitialized DenseUNet-{name} (pretrained={pretrained}).")

    def forward(self, x, dropout=True):
        f = F.relu(self.features(x))  # 32x downsampled -> ReLU (is BN'd)
        x = self.up1(self.encoder_hooks[3].features, f)  # cat block3 out -> 16x down
        x = self.up2(self.encoder_hooks[2].features, x)  # cat block2 out -> 8x down
        x = self.up3(self.encoder_hooks[1].features, x)  # cat block1 out -> 4x down
        x = self.up4(self.encoder_hooks[0].features, x)  # cat activated 7x7 out -> 2x

        if self.use_deconv:
            x_fea = self.final_deconv(x)
        else:
            x_fea = F.interpolate(x, scale_factor=2, mode='bilinear', 
                align_corners=True)
        x_fea = self.conv1(x_fea)
        
        if dropout:
            x_fea = F.dropout2d(x_fea, p=0.3)
        
        x_fea = F.relu(self.bn1(x_fea))
        x_out = self.conv2(x_fea)
        
        # Classifier
        c_out = F.relu(f, inplace=True)
        c_out = F.avg_pool2d(c_out, kernel_size=7, stride=1).view(c_out.shape[0], -1)
        c_out = F.dropout(c_out, p=0.2, training=self.training)
        c_out = self.classifier(c_out)
        return f, c_out, x_out

    def close(self):
        for hook in self.encoder_hooks: 
            hook.remove()

    def _setup_encoder(self, name, pretrained=True):
        if '121' in name:
            base_model = densenet.densenet121
            layer_dims = densenet_outdims['densenet121']
        elif '169' in name:
            base_model = densenet.densenet169
            layer_dims = densenet_outdims['densenet169']
        elif '201' in name:
            base_model = densenet.densenet201
            layer_dims = densenet_outdims['densenet201']
        elif '161' in name:
            base_model = densenet.densenet161 
            layer_dims = densenet_outdims['densenet161']
        else:
            raise Exception(f"Invalid DenseNet name: {name}.")

        # layers is a list of 2 modules: dnet.features, dnet.classifier
        comps = list(base_model(pretrained=pretrained,
                        num_classes=1).children())
        layers = nn.Sequential(*comps)  # 0: feat extractor, 1: lin classifier
        
        encoder_hooks = [
            ForwardHook(layers[0][2]),  # ReLU after 7x7
            ForwardHook(layers[0][4]),  # DenseBlock 1 out
            ForwardHook(layers[0][6]),  # DenseBlock 2 out
            ForwardHook(layers[0][8]),  # DenseBlock 3 out
        ]
        return layers[0], encoder_hooks, layer_dims




""" Module rere/data/loading/samplers.py

Image samplers are used in the dataloading process to facilitate creation
of batch elements, especially when a single sample produces multiple examples.

Common use cases of samplers include:
 1. Patch sampling (N patches per volume) in medical image segmentation.
 2. Inference sampling where multiple transforms are applied to an image
     during inference time (samplers coupled with an aggregator).
 3. During 2D self-supervised training when a single image produces 2 large
     crops as well as additional tiny crops (e.g. in SwAV).
"""

import collections
from abc import ABC, abstractmethod
import numpy as np
import torch
import nibabel as nib

import dmt
from dmt.data import Sample
from dmt.transforms import ImageHarmonizer

from dmt.utils.parse import parse_bool, parse_positive_int


class PatchSampler(ABC):
    """ Abstract class & base functionality definition for patch samplers.
    
    Implmentation Notes:
        - Any child class of this base class must implement self.ndim for 2D/3D
        - __init__ always only takes patch_size
        - __call__ is pretty much the only API of this class. It returns a
            generator of patches in the input image format you gave. 
    """
    
    def __init__(self, patch_size, return_sample=False, copy=True):
        """
        Args:
            patch_size: a sequence indicating the patch_size to crop
                Can be (1) single int or sequence of length 1
                       (2) a sequence of appropriate dimensionality
            return_sample: flag that returns a sample if True, otherwise it
                returns an image of the same type as the one given to __call__.
            copy: flag that copies the harmonizer's sample after every get
        """
        self.patch_size = self._parse_patch_size(patch_size, self.ndim)
        self.return_sample = parse_bool(return_sample, 'return_sample')
        self.copy = parse_bool(copy, 'copy')

    def __call__(self, data, num_patches=None, include_keys=None):
        """ Main API for sampler calls. Returns a generator of crops. 
        Args:
            data: Can be a Sample, Image, Harmonizer, Torch tensor, Numpy array, 
                or SimpleITK Image.
            num_patches: number of patches the iterator will return. If it's
                None or <= 0, an infinite iterator will be returned.
            include_keys: Only used if a sample or dict is given. Only images
                with keys in include_keys will be cropped. If None, all images
                will be cropped. 
        Returns:
            Iterator of transformed image(s) in the original type given. 
        """
        
        include_keys, _ = Sample._parse_include_exclude_keys(include_keys, None)
        harmonizer = ImageHarmonizer(data, include_keys=include_keys, 
                                     copy=self.copy)

        # make sure crop size is smaller than all relevant images
        images_shape = harmonizer.consistent_shape
        if images_shape is None or np.any(self.patch_size > images_shape):
            msg = (f'Patch size {tuple(self.patch_size)} cannot be'
                   f' larger than image size {tuple(images_shape)}')
            raise RuntimeError(msg)
        return self._generate_patches(harmonizer, num_patches=num_patches)
        
    
    @abstractmethod
    def _generate_patches(self, sample, num_patches=None):
        """ Implement for all PatchSampler inheritors. 
        Returns a generator of patches to __call__(). """
        pass
    
    def crop(self, harmonizer, lower_indices, patch_size):
        upper_indices = lower_indices + patch_size
        
        # Crop Transform
        sample = harmonizer.sample  # copies data automatically
        for k in harmonizer.include_keys:
            image = sample[k]
            cropim = self.crop_dmt_image(image, lower_indices, upper_indices)
            sample[k] = cropim
            
        # Record information
        location = lower_indices.tolist() + upper_indices.tolist()
        sample.record_transform({
            'name': 'Crop',
            'p': 100,
            'include_keys': harmonizer.include_keys,
            'location': location
        })
        if self.return_sample:
            return sample
        crop_output = harmonizer.get_output(sample)
        return crop_output
    
    def _parse_patch_size(self, patch_size, ndim):
        msg = ('"patch_size" must be an int or a sequence of ints, given '
               f'{type(patch_size)}')
        assert isinstance(patch_size, (int, collections.Sequence)), msg
        
        patch_size_list = []
        if isinstance(patch_size, int):
            patch_size = parse_positive_int(patch_size, 'patch_size')
            patch_size_list = [patch_size] * ndim
        else:
            patch_size = list(patch_size)
            msg = ('If "patch_size" is a sequence, it must be either length '
                   f'1 or 3. Given: length {len(patch_size)}')
            assert len(patch_size) in (1, ndim), msg
            
            if len(patch_size) == 1:
                patch_size = [patch_size[0]] * ndim
                
            for size in patch_size:
                size = parse_positive_int(size, 'patch_size number')
                patch_size_list.append(size)
                
        assert len(patch_size_list) == ndim, 'Sanity gone.'
        return np.array(patch_size_list).astype(np.uint16)
    
    def __repr__(self):
        string = (
            f"{type(self).__name__}\n"
            f"  Patch Size = {self.patch_size} "
        )
        return string


class RandomSampler(PatchSampler):
    
    @abstractmethod
    def get_probability_map(self, harmonizer):
        raise NotImplementedError


##### ----- ### ------ #  Patch Dimension Base Classes   # ----- ### ----- #####

class PatchSamplerScalar3D(PatchSampler):
    ndim = 3
    
    def crop_dmt_image(self, image, lower_indices, upper_indices):
        new_origin = nib.affines.apply_affine(image.affine, lower_indices)
        new_affine = image.affine.copy()
        new_affine[:3, 3] = new_origin
        
        i0, j0, k0 = lower_indices
        i1, j1, k1 = upper_indices
        cropped_sitk_image = image.image[i0:i1, j0:j1, k0:k1]
        
        from dmt.data import ScalarImage3D
        new_image = ScalarImage3D(cropped_sitk_image)
        # print(image.origin, new_origin, '\n\n', image.affine, new_affine)
        new_image.origin = new_origin
        new_image.affine = new_affine
        return new_image
    
    
class PatchSamplerScalar2D(PatchSampler):
    ndim = 2
    
    def crop_dmt_image(self, image, lower_indices, upper_indices, affine):
        raise NotImplementedError()
    
class PatchSamplerVector2D(PatchSampler):
    ndim = 2
    
    def crop_dmt_image(self, image, lower_indices, upper_indices, affine):
        raise NotImplementedError()
    
    
##### ----- ### ------ #    Sampler Type Base Classes    # ----- ### ----- #####

    
def crop_dmt_image(image, lower_indices, upper_indices, affine):
    affine = image.affine.copy()
    new_origin = nib.affines.apply_affine(image.affine, lower_indices)
    new_affine = image.affine.copy()
    new_affine[:3, 3] = new_origin
    
    i0, j0, k0 = lower_indices
    i1, j1, k1 = upper_indices
    cropped_image = image[:, i0:i1, j0:j1, k0:k1]
    cropped_image.origin = new_origin
    cropped_image.affine = new_affine
    return cropped_image


    




__all__ = [
    'Image',
    'ScalarImage3D',
    'ScalarImage2D',
    'VectorImage2D',
    'Label',
    'CategoricalLabel',
    'ScalarMask2D',
    'Scalarmask3D',
    'Sample',
    'SampleSet',
    'OneToOneLoader',
    'OneToManyLoader',
]

# Sample Components: images, labels, categories
from .image_base import Image
from .images import (
    ScalarImage3D,
    ScalarImage2D,
    VectorImage2D,
)

from .label_base import Label
from .label_categories import CategoricalLabel
from .label_masks import (
    ScalarMask3D,
    ScalarMask2D
)

# Samples & their collections
from .samples.sample import Sample
from .samples.sampleset import SampleSet

# Data Loaders
from .loading.oto_loader import OneToOneLoader
from .loading.otm_loader import OneToManyLoader

from .loading.patches.uniform import UniformSampler3D




""" Module metrics.py (By: Charley Zhang, Nov 2020)

Classification metrics:
  Note: these are computed for each class. You must average the results.
    - Confusion matrix (TP, FP, FN, TN)
    - AUCs
    - Specificities
    - Recalls
    - Precisions
    - F1s
    - Accuracies
    - Balanced Accuracies
"""

import torch
import numpy as np
from sklearn.metrics import roc_auc_score
from sklearn.metrics import confusion_matrix, balanced_accuracy_score
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score

import pdb

# Turn off sklearn warnings
import warnings
warnings.filterwarnings(action='ignore')


### ======================================================================== ###
### * ### * ### * ### *      Metrics Bundle Coverage     * ### * ### * ### * ###
### ======================================================================== ###


def multiclass_metrics(pred, targ):
    """ IPMI: classification for CIFAR & ISIC.
    Metrics: AUC, F1, acc, sens, spec.
    Parameters:
        pred (NxC tensor) - per class probabilities
        targ (NxC tensor) - one-hot ground truth
    """
    def to_np(tens):
        return tens.float().cpu().detach().numpy()
    
    targ = to_np(targ)
    pred_1hot = to_np(pred.scatter_(1, pred.argmax(1).view(-1,1), 1)).astype('uint8')
    pred_prob = to_np(pred)

    AUCs = compute_aucs(pred_prob, targ)
    ACCs = compute_accuracies(pred_1hot, targ)
    F1s  = compute_f1s(pred_1hot, targ)
    SENs = compute_recalls(pred_1hot, targ)
    bal_ACC = compute_balanced_accuracies(pred_1hot.argmax(1), targ.argmax(1))
    SPEs, Ss = compute_specificities(pred_1hot, targ)
    
    # Ss = (tns, fps, fns, tps)
    TNs, FPs, FNs, TPs = Ss
    overall_f1 = ((2 * TPs) / (2 * TPs + FPs + FNs)).mean()

    return {
        'AUCs': AUCs,
        'ACCs': ACCs,
        'F1s': F1s,
        'SENs': SENs,
        'SPEs': SPEs,
        'TNs': TNs,
        'FPs': FPs,
        'FNs': FNs,
        'TPs': TPs,
        'oF1': overall_f1,
        'F1': F1s.mean(),
        'AUC': AUCs.mean(),
        'ACC': ACCs.mean(),
        'SEN': SENs.mean(),
        'SPE': SPEs.mean(),
        'bACC': bal_ACC
    }


def multilabel_metrics(pred, targ, thresh=0.5):
    pred_1hot = pred >= thresh
    pass


### ======================================================================== ###
### * ### * ### * ### * Per Class Classification Metrics * ### * ### * ### * ###
### ======================================================================== ###


def compute_confusion_matrix(pred, targ, C):
    """ Compute per class classification precision scores.
    Parameters:
        pred (Nx1 tensor or np) - prediction one-hot
        targ (Nx1 tensor or np) - target one-hot
        C - num_classes
    Returns:
        tuple(TN, FP, FN, TP)
    """
    assert targ.shape == pred.shape, f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    if pred.ndim == 2 and pred.shape[1] > 1:
        pred = pred.argmax(1)
        targ = targ.argmax(1) 
    tn, fp, fn, tp = confusion_matrix(targ, pred, labels=range(C)).ravel()
    return tn, fp, fn, tp


def compute_aucs(pred, targ):
    """
    Computes Area Under the Curve (AUC) from prediction scores.
    Args:
        targ: Pytorch tensor on GPU, shape = [n_samples, n_classes]
          true binary labels.
        pred: Pytorch tensor on GPU, shape = [n_samples, n_classes]
          can either be probability estimates of the positive class,
          confidence values, or binary decisions.
    Returns:
        List of AUROCs of all classes.
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    AUROCs = np.zeros(targ.shape[-1])
    for i in range(targ.shape[-1]):
        try:
            AUROCs[i] = roc_auc_score(targ[:, i], pred[:, i])
        except ValueError:
            AUROCs[i] = 0
    return AUROCs


def compute_specificities(pred, targ):
    """ Compute per class specificity scores: TN/(TN+FP)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    dims = targ.shape[-1]
    specs = np.zeros(dims)
    tns, fps, fns, tps = np.zeros(dims), np.zeros(dims), np.zeros(dims), np.zeros(dims)
    for i in range(targ.shape[-1]):
        try:
            tn, fp, fn, tp = compute_confusion_matrix(targ[:, i], (pred[:, i]), 2)
            specs[i] = tn / (tn + fp + 1e-7)
            tns[i] = tn
            fps[i] = fp
            fns[i] = fn
            tps[i] = tp
        except ValueError as error:
            print('Error in computing specificity for {}.\n Error msg:{}'.format(i, error))
            specs[i] = 0
    return specs, (tns, fps, fns, tps)


def compute_recalls(pred, targ):
    """ Compute per class classification recall/sensitivity scores: TP/(TP+FN)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    recalls = np.zeros(targ.shape[-1])
    for i in range(targ.shape[-1]):
        try:
            recalls[i] = recall_score(targ[:, i], (pred[:, i]))
        except ValueError as error:
            print('Error in computing recall for {}.\n Error msg:{}'.format(i, error))
            recalls[i] = 0
    return recalls


def compute_precisions(pred, targ):
    """ Compute per class classification precision scores: TP/(TP+FP)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    precisions = np.zeros(targ.shape[-1])
    for i in range(targ.shape[-1]):
        try:
            precisions[i] = recall_score(targ[:, i], (pred[:, i]))
        except ValueError as error:
            print('Error in computing precision for {}.\n Error msg:{}'.format(i, error))
            precisions[i] = 0
    return precisions


def compute_f1s(pred, targ):
    """ Compute per class classification F1 scores: 2*TP/(2*TP+FP+FN)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    f1s = np.zeros(targ.shape[-1])
    for i in range(targ.shape[-1]):
        try:
            f1s[i] = f1_score(targ[:, i], (pred[:, i]))
        except ValueError as error:
            print('Error in computing f1 for {}.\n Error msg:{}'.format(i, error))
            f1s[i] = 0
    return f1s


def compute_accuracies(pred, targ):
    """ Compute per class classification accuracies: (TP+TN)/(TP+TN+FP+FN)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape[1] == pred.shape[1], f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    accuracies = np.zeros(targ.shape[-1])
    for i in range(targ.shape[-1]):
        try:
            accuracies[i] = accuracy_score(targ[:, i], (pred[:, i]))
        except ValueError as error:
            print('Error in computing accuracy for {}.\n Error msg:{}'.format(i, error))
            accuracies[i] = 0
    return accuracies


def compute_balanced_accuracies(pred, targ):
    """ Compute per class classification accuracies: (TP+TN)/(TP+TN+FP+FN)
    Parameters:
        pred (NxC tensor or np) - prediction one-hot
        targ (NxC tensor or np) - target one-hot
    """
    assert targ.shape == pred.shape, f"pred:{pred.shape} targ:{targ.shape}"

    if isinstance(pred, torch.Tensor):
        pred = pred.float().cpu().detach().numpy()
    if isinstance(targ, torch.Tensor):
        targ = targ.float().cpu().detach().numpy()
    assert isinstance(pred, np.ndarray), 'pred must be tensor or np array'
    assert isinstance(targ, np.ndarray), 'targ must be tensor or np array'

    balanced_acc = balanced_accuracy_score(targ, pred)
    return balanced_acc




### Sanity Checks

if __name__ == '__main__':
    # Multi-Class (4x3 tensors)
    pred = torch.tensor([[0.1485, 0.4298, 0.4216],
                         [0.4826, 0.4813, 0.0361],
                         [0.6463, 0.0945, 0.2593],
                         [0.1980, 0.5257, 0.2763]])
    pred_oh = torch.tensor([[0., 1., 0.],
                            [1., 0., 0.],
                            [1., 0., 0.],
                            [0., 1., 0.]])
    targ = torch.tensor([[0., 1., 0.],
                         [0., 1., 0.],
                         [0., 1., 0.],
                         [0., 1., 0.]])

    aucs = compute_aucs(pred, targ)
    accs = compute_accuracies(pred_oh, targ)
    f1s = compute_f1s(pred_oh, targ)
    sens = compute_recalls(pred_oh, targ)
    specs = compute_specificities(pred_oh, targ)
    import IPython; IPython.embed(); 




""" Module __init__.py for dmt.metrics """

from dmt.metrics import medpy_metrics


### ------ #     Classification Metrics     # ----- ###
confusion_matrix = None
area_under_curve = None
f1_score = None

### ------ #     Localization Metrics     # ----- ###

### ------ #     Binary Array Surface Metrics     # ----- ###
dice_2d = dice_3d = medpy_metrics.dc
jaccard_2d = jaccard_3d = medpy_metrics.jc
hausdorff_2d = hausdorff_3d = medpy_metrics.hd
hausdorff95_2d = hausdorff_3d = medpy_metrics.hd95
precision_2d = precision_3d = medpy_metrics.precision
recall_2d = recall_3d = medpy_metrics.recall
true_positive_rate_2d = true_positive_rate_3d = medpy_metrics.true_positive_rate
true_negative_rate_2d = true_negative_rate_3d = medpy_metrics.true_negative_rate
sensitivity_2d = sensitivity_3d = medpy_metrics.sensitivity
specificity_2d = specificity_3d = medpy_metrics.specificity
positive_predicitive_value_2d = medpy_metrics.positive_predictive_value
positive_predicitive_value_3d = medpy_metrics.positive_predictive_value
average_surface_distance_2d = average_surface_distance_3d = medpy_metrics.asd
average_symmetric_surface_distance_2d = medpy_metrics.assd
average_symmetric_surface_distance_3d = medpy_metrics.assd
relative_absolute_distance_2d = medpy_metrics.ravd
relative_absolute_distance_3d = medpy_metrics.ravd




""" Module losses3d.py (By: Charley Zhang, 2021.05)
Contains loss functions specifically for 3D imaging tasks.

Assumes the following:
 - Prediction & target variables are tensors.
 - Prediction & target variables have the same shape (target = one-hot format)
"""


import torch
import torch.nn as nn
import torch.nn.functional as F


REDUCTIONS = ['mean', 'sum', 'class', 'none']


class DiceLoss3d:
    """ Good for both 2D and 3D tensors. Assumes pred.shape == targ.shape. 
    Whether it is soft-dice or dice depends on whether predictions are 
     thresholded to be a binary mask (dice) or as probabilities (soft-dice).
    """        
    def __init__(self, smooth=1.):
        self.smooth = smooth
    
    def __call__(self, pred, targ):
        return dice_loss(pred, targ, s=self.smooth)


class CrossEntropyLoss3d:
    """ Different from torch CE in 2 ways: (1) assumes pred.shape == targ.shape, 
        which means targ must be one-hot, (2) does not apply softmax to pred.
    """
    def __init__(self, weights=None, reduction='mean'):
        self.weights = weights
        self.reduction = reduction
    
    def __call__(self, pred, targ):
        return cross_entropy_loss(pred, targ, weights=self.weights,
                                  reduction=self.reduction)


class DiceCrossEntropyLoss3d:
    def __init__(self, weights=None, alpha=0.5):
        self.weights = weights
        self.alpha = alpha
    
    def __call__(self, pred, targ):
        ce_loss = cross_entropy_loss(pred, targ, weights=self.weights,
                                  reduction='mean')
        dice_loss = dice_loss(pred, targ)
        return self.alpha * dice_loss + (1 - self.alpha) * ce_loss
    

        

### ======================================================================== ###
### * ### * ### * ### *           Functional             * ### * ### * ### * ###
### ======================================================================== ###


def dice_loss(pred, targ, s=1):
    """ Dice loss. Assumes targ is in one-hot format. 
    Parameters
        pred - prediction image probabilities [any shape]
        targ - binary label image [any shape]
        s (float) - smoothing factor added to the numerator and denominator.
    """
    assert pred.shape == targ.shape 
    pred_flat = pred.view(-1).float()
    targ_flat = targ.view(-1).float()
    intersection = (pred_flat, targ_flat).sum()
    dice = (2 * intersection + s) / (pred_flat.sum() + targ_flat.sum() + s)
    return 1 - dice


def cross_entropy_loss(pred, targ, weights=None, reduction='mean'):
    """ CE loss w/probabilties. Assumes targ is in one-hot format. 
    Parameters
        pred - prediction probabilities [BxCxHxWxD]
        targ - binary label image [BxCxHxWxD]
        s (float) - smoothing factor added to the numerator and denominator.
    """
    if pred.shape != targ.shape:
        assert targ.shape[0] == pred.shape[0]
        assert targ.shape[1:] == pred.shape[2:]
    assert reduction in REDUCTIONS, f'Reduction {reduction} is not valid!'
    
    targ_ind = to_single_index(targ, keepdims=False)  # BxHxWxD
    loss = F.nll_loss(torch.log(pred), targ_ind, weight=weights, 
                      reduction=reduction)
    return loss
    

### ======================================================================== ###
### * ### * ### * ### *       Rudimentary Testing        * ### * ### * ### * ###
### ======================================================================== ###


def to_one_hot(target, C=None):
    """ 
    Parameters
        tensor (torch.Tensor) - Input of shape Bx1xHxWxD
        C (int) - Number of classes
            Note: if C is none, max(unique_values) is used
    """
    assert target.ndim == 5, 'Expected 5 dimensional input (Bx1xHxWxD)'
    assert target.shape[1] == 1, 'Target shape needs to be Bx1xHxWxD'
    if C is None:
        C = target.max() + 1
    one_hot_shape = (target.shape[0], C, *target.shape[2:])
    one_hot = torch.zeros(one_hot_shape, device=target.device)
    one_hot.scatter_(1, target, 1)
    return one_hot

def to_single_index(target, keepdims=True):
    """
    Parameters
        target (tensor) - one-hot target BxCxHxWxD
        keepdims (bool) - if True, returns Bx1xHxWxD else BxHxWxD
    """
    assert target.ndim == 5, 'Expected 5 dimensional input (BxCxHxWxD)'
    targ_ind = target.view(target.shape[0], target.shape[1], -1).argmax(1)
    targ_ind = targ_ind.view(target.shape[0], *target.shape[2:])
    if keepdims:
        targ_ind = targ_ind.unsqueeze(1)
    return targ_ind
    
if __name__ == '__main__':
    
    ### Cross Entropy ###
    print(f"Testing CE ..", end='')
    C = 4
    pred = torch.randn(3, C, 1, 2, 3)
    targ = torch.randint(0, C, (3, 1, 1, 2, 3))
    assert torch.all(targ == to_single_index(to_one_hot(targ)))  
    
    torch_ce = nn.CrossEntropyLoss()
    torch_loss = torch_ce(pred, targ.squeeze(1))
    
    targ_1h = to_one_hot(targ)
    loss = cross_entropy_loss(pred.softmax(1), targ_1h, reduction='mean')
    assert abs(torch_loss - loss) < 10**-6, \
           f'Torch {torch_loss.item():.4f}, Ours {loss.item():.4f}.'
    print(f"good ✔")
    
    # Now weight weights
    print(f"Testing CE w/class-weights..", end='')
    pred = torch.randn(3, C, 1, 2, 3)
    targ = torch.randint(0, C, (3, 1, 1, 2, 3))
    weights = torch.randint(1, 3, (C,)).float()
    assert torch.all(targ == to_single_index(to_one_hot(targ)))  
    
    for reduction in ('none', 'sum', 'mean'):
        print(f' reduction {reduction}..', end='')
        torch_ce = nn.CrossEntropyLoss(weight=weights, reduction=reduction)
        torch_loss = torch_ce(pred, targ.squeeze(1))

        targ_1h = to_one_hot(targ)
        # import IPython; IPython.embed(); 
        loss = cross_entropy_loss(pred.softmax(1), targ_1h, reduction=reduction,
                                  weights=weights)
        
        assert torch.all((torch_loss - loss) < 10**-6), \
               f'Torch {torch_loss:.4f}, Ours {loss:.4f}.'
    print(f"good ✔")
    
    



import numbers
import collections
import pprint
import torch
import numpy as np

from .label_base import Label


class CategoricalLabel(Label):
    """ Class for categorical labels. Shared functionality definition. 
    Implementation Details:
        - Internally, labels are stored as a list of class_ids with a class's
            id defined by its index in the 'class_names' argument
    """
    
    def __init__(
            self, 
            label, 
            class_names, 
            multi_label=False,
            **kwargs
            ):
        """
        Args:
            label: the categor(y/ies) of label. Can be a number (class id), a 
                string (class name), a sequence of numbers (class ids), or a 
                sequence of strings (class names). 
            class_names: a sequence of strings indicating the names of classes
            multi_label: flag for if it is a multi-label task 
            **kwargs: additional custom attributes to be added to label
        """
        
        class_names = self._parse_class_names(class_names)
        self._class_names = class_names
        
        self._multi_label = multi_label
        self._given_label = label
        self._ids = self._parse_label(label, multi_label, class_names)
        
        super().__init__(**kwargs)  # store extra custom attributes 
    
    @property
    def class_names(self):
        return self._class_names
    
    @property
    def num_classes(self):
        return len(self._class_names)
    
    @property
    def unique_values(self):
        return self._ids
    
    def get_one_hot(self, tensor=False):
        ids = self.get_ids(squeeze=False)
        oh = np.zeros(self.num_classes, dtype=np.uint8)  # uint8, max val = 1
        for i in ids:
            oh[i] = 1
        
        if tensor:
            oh = torch.tensor(oh).double()
        return oh
    
    def get_ids(self, tensor=False, squeeze=True):
        ids = self._ids
        if tensor:
            ids = torch.tensor(ids).double()
        if squeeze:
            ids = ids.squeeze()
        return ids
    
    def _parse_class_names(self, class_names):
        # Verify if given class_names is valid
        cn_type = type(class_names)
        msg = f'Argument "class_names" must be a sequence. Given: {cn_type}.'
        assert isinstance(class_names, collections.Sequence), msg
        
        try:
            class_names = [str(cn) for cn in list(class_names)]
        except:
            msg = ('Argument "class_names" must be a sequence where each '
                   'element can be represented as a string.')
            raise ValueError(msg)
            
        return class_names
    
    def _parse_label(self, label, multi_label, class_names):
        if isinstance(label, (numbers.Number, str)):
            label = [label]
        
        l_type = type(label)
        msg = (f'Argument "label" must be a number, a string, or a squence of '
               f'either. You gave: {l_type}.')
        assert isinstance(label, collections.Sequence), msg
        
        assert len(label) > 0, f'Given label sequence cannot be empty.'
        
        n_classes = len(class_names)
        ids = []
        for lab in label:
            if isinstance(lab, int):  # assumed to be the class id
                msg = (f'Given class id "{lab}" in labels "{label}" is larger '
                       f'than #classes - 1.')
                assert lab < n_classes, msg
                
                ids.append(lab)
                continue
            elif isinstance(lab, float):  # assumed to be class name
                lab = str(lab)
                
            if isinstance(lab, str):  # assumed to be class name
                msg = (f'Give class name "{lab}" in labels "{label}" is not in '
                       f'class_names ({class_names}).')
                assert lab in class_names, msg
                ids.append(class_names.index(lab))
            else:
                l_type = type(lab)
                msg = (f'Argument "label" must contain a number (class id), or '
                       f'a string (class name). You gave: {l_type}.')
                raise ValueError(msg)
                
        if not multi_label:
            msg = (f'You gave multiple categories in a multi-class task. '
                   'Set init argument "multi_label" to be true.')
            assert len(ids) <= 1, msg
        
        return np.array(sorted(ids), dtype=np.int32)
    
    def __repr__(self):
        cat_type = 'multi-label' if self._multi_label else 'multi-class'
        ids = list(self.get_ids(squeeze=False))
        lab_names = [self.class_names[i] for i in ids]
        attrs = {k: v for k, v in self.items() if k not in self._default_attrs}
        attrs_str = pprint.pformat(attrs, indent=2, compact=True)
        string = (
            f'{self.__class__.__name__} ({cat_type}, gid={self.gid})\n'
            f'  Label: cls-ids={ids}, cls-names={lab_names}\n'
            f'  {self.num_classes} classes: {self.class_names}\n'
            f'Other Attributes: \n {attrs_str} \n')
        return string
            



""" module contrastive.py
Sources:
> https://github.com/sthalles/SimCLR/blob/master/loss/nt_xent.py

"""

import torch
import numpy as np


class NTXentLoss(torch.nn.Module):

    def __init__(self, device, batch_size, temp=0.5, use_cosine=True):
        super(NTXentLoss, self).__init__()
        self.batch_size = batch_size
        self.temperature = temp
        self.device = device
        self.softmax = torch.nn.Softmax(dim=-1)
        self.mask_samples_from_same_repr = self._get_correlated_mask().type(torch.bool)
        self.similarity_function = self._get_similarity_function(use_cosine)
        self.criterion = torch.nn.CrossEntropyLoss(reduction="sum")

    def _get_similarity_function(self, use_cosine_similarity):
        if use_cosine_similarity:
            self._cosine_similarity = torch.nn.CosineSimilarity(dim=-1)
            return self._cosine_simililarity
        else:
            return self._dot_simililarity

    def _get_correlated_mask(self):
        diag = np.eye(2 * self.batch_size)
        l1 = np.eye((2 * self.batch_size), 2 * self.batch_size, k=-self.batch_size)
        l2 = np.eye((2 * self.batch_size), 2 * self.batch_size, k=self.batch_size)
        mask = torch.from_numpy((diag + l1 + l2))
        mask = (1 - mask).type(torch.bool)
        return mask.to(self.device)

    @staticmethod
    def _dot_simililarity(x, y):
        v = torch.tensordot(x.unsqueeze(1), y.T.unsqueeze(0), dims=2)
        # x shape: (N, 1, C)
        # y shape: (1, C, 2N)
        # v shape: (N, 2N)
        return v

    def _cosine_simililarity(self, x, y):
        # x shape: (N, 1, C)
        # y shape: (1, 2N, C)
        # v shape: (N, 2N)
        v = self._cosine_similarity(x.unsqueeze(1), y.unsqueeze(0))
        return v

    def forward(self, zis, zjs):
        representations = torch.cat([zjs, zis], dim=0)

        similarity_matrix = self.similarity_function(representations, representations)

        # filter out the scores from the positive samples
        l_pos = torch.diag(similarity_matrix, self.batch_size)
        r_pos = torch.diag(similarity_matrix, -self.batch_size)
        positives = torch.cat([l_pos, r_pos]).view(2 * self.batch_size, 1)

        negatives = similarity_matrix[self.mask_samples_from_same_repr].view(2 * self.batch_size, -1)

        logits = torch.cat((positives, negatives), dim=1)
        logits /= self.temperature

        labels = torch.zeros(2 * self.batch_size).to(self.device).long()
        loss = self.criterion(logits, labels)

        return loss / (2 * self.batch_size)